__license__ = 'LGPL-3.0-or-later'
__copyright__ = 'Copyright 2024  W. Braun (epiray GmbH)'
__authors__ = 'P. Bredol'
__url__ = 'https://github.com/zaphB/freecad.optics_design_workbench'

try:
  import FreeCADGui as Gui
  import FreeCAD as App
except ImportError:
  pass

import subprocess
import os
import sys
import signal

from ... import io
from .. import results_store
from .. import processes

_WORKER_INDEX = 0

class WorkerProcess:
  '''
  This class represents one background process worker. It needs to known which
  simulation type is going on and which result folder to use and will run the 
  simulation_loop on its own.
  '''

  def __init__(self, simulationType, simulationRunFolder):
    # set index for worker process for easy identification in cli logs
    global _WORKER_INDEX
    self.index = _WORKER_INDEX
    _WORKER_INDEX += 1

    # start freecad in cli mode (-c) with current document as active document
    self.simulationType = simulationType
    self.simulationFilePath = os.path.realpath(processes.simulatingDocument().getFileName())
    self.simulationRunFolder = simulationRunFolder
    self._isRunning = True

    # try to extract freecad executable path: first check APPIMAGE environment variable,
    # to find out if running appimage, then try executable found in sys.executable,
    # if that does not look like a freecad executable let shell decide
    if freecadPath := os.environ.get('APPIMAGE', None):
      pass
    elif 'freecad' in sys.executable.lower():
      freecadPath = sys.executable
    else:
      freecadPath = 'FreeCAD'
    io.verb(f'detected freecad executable "{freecadPath}"')

    # launch child process
    self._p = subprocess.Popen([freecadPath, '-c', self.simulationFilePath],
                                stdout=subprocess.DEVNULL,
                                stderr=subprocess.DEVNULL,
                                stdin=subprocess.PIPE, 
                                text=True, bufsize=-1)

    # write python snippet to start desired simulation mode
    self.say('entering simulation loop...')
    self.write(f'\r\n'
               f'import freecad.optics_design_workbench.simulation\r\n'
               f'freecad.optics_design_workbench.simulation.runSimulation('
                      f'action="{self.simulationType}", '
                      f'slaveInfo=dict(simulationRunFolder="{self.simulationRunFolder}", '
                      f'               parentPid={os.getpid()}))\r\n'
               f'exit()\r\n')

    self._isquit = False
    self._isterminate = False
    self._iskill = False

  def write(self, data):
    self._p.stdin.write(data)
    self._p.stdin.flush()

  def say(self, msg):
    if type(msg) is not str:
      msg = bytes(msg).decode('utf8')
    for line in msg.split('\n'):
      line = line.strip().strip('>').strip()
      if len(line):
        io.info(f'worker ({self.index}) says: '+line)

  def isRunning(self):
    if self._isRunning:
      if (res:=self._p.poll()) is not None:
        self._isRunning = False
        self.say(f'finished (exit code {res})')
    return self._isRunning

  def quit(self):
    if self.isRunning() and not self._isquit:
      self.say('asking FreeCAD to quit...')
      self._isquit = True

  def terminate(self):
    if self.isRunning():
      if not self._isterminate:
        self.say('terminating FreeCAD...')
        self._isterminate = True
      try:
        self._p.stdin.close()
      except:
        pass
      self._p.send_signal(signal.SIGTERM)

  def kill(self):
    if self.isRunning():
      if not self._iskill:
        self.say('killing FreeCAD...')
        self._iskill = True
      self._p.send_signal(signal.SIGKILL)



'''
Classes to generate random numbers according to arbitrary probability distribution functions.
'''

__license__ = 'LGPL-3.0-or-later'
__copyright__ = 'Copyright 2024  W. Braun (epiray GmbH)'
__authors__ = 'P. Bredol'
__url__ = 'https://github.com/zaphB/freecad.optics_design_workbench'

# make sure bool() is not overloaded by numpy import
_bool = bool
from numpy import *
bool = _bool

import sympy as sy
import time
import signal

from . import points_by_density
from .. import io


def _setAlarm(deadline):
    timeout = deadline-time.time()
    # it may seem a bit drastic to raise a KeyboardInterrupt here, but other Exceptions
    # seem to be caught by sympy internally in certain cases. Only KeyboardInterrupt
    # is able to actually stop a hung sympy function.
    if timeout < 0:
      raise KeyboardInterrupt('time is up')
    def handler(sig, n):
      raise KeyboardInterrupt('time is up')
    signal.signal(signal.SIGALRM, handler)
    signal.alarm(int(timeout)+1)


def _clearAlarm():
  signal.alarm(0)


class VectorRandomVariable:
  '''
  Vector valued random variable. 
  '''
  def __init__(self, probabilityDensity, variableDomains={}, numericalResolutions={}, variableOrder=None,
               warnIfDiscretizationStepAbove=5e-2):
    self._probabilityDensity = probabilityDensity
    self._probabilityDensityBaseExpr = None
    self._variables = None
    self._variableDomains = variableDomains
    self._numericalResolutions = numericalResolutions
    self._variableOrder = variableOrder
    self._constantsDict = {}
    self._mode = 'not yet compiled'
    self._warnIfDiscretizationStepAbove = warnIfDiscretizationStepAbove


  def compile(self, timeout=2, disableAnalytical=False, **kwargs):
    '''
    Draw random numbers (or vectors) following the distribution represented by this object.

    Arguments:
    ==========
    timeout : numeric
      Give up on finding an analytical solution and fallback to numeric mode after timeout seconds.

    disableAnalytical : boolean
      Set to true to completely disable analytic mode.

    **kwargs    
      Dictionary of constants to substitute in the distribution expression.
    '''
    self._deadline = time.time()+timeout
    self._setConstants(**kwargs)

    try:
      # immediately go to numerical fallback if analytical is disabled
      if disableAnalytical:
        raise ValueError('stop')

      # try analytical treatment first
      self._transformLambdas = [self._generateAnalyticScalarLambda(i) for i in range(len(self._variables))]
      self._mode = 'analytic'

    # fallback to numerical treatment and analytical mode did not succeed
    except Exception:
      self._transformLambdas = [self._generateNumericScalarLambda(i) for i in range(len(self._variables))]
      self._mode = 'numeric'


  def mode(self):
    return self._mode


  def showExpressions(self, simplify=True):
    '''
    Pretty print generated internal expressions and lambdas for debugging purposes.
    '''
    print('probability density expression: ', self._probabilityDensityExpr, ' variables: ', self._variables)
    print(self._transformLambdas)
    for i, var in enumerate(self._variables):
      print(f'variable "{var}" '+('conditional' if i<len(self._variables)-1 else '')+f' probability density: ')
      probDens, integral, invertedSols = self._transformLambdas[i][0]._origExpressions
      if simplify and str not in [type(x) for x in (probDens, integral, invertedSols)]:
        probDens = probDens.simplify()
        integral = integral.simplify()
        invertedSols = [sol.simplify() for sol in invertedSols]
      print(f'  conditional prop. dens.: ', probDens)
      print(f'  integrated prop. dens.: ', integral)
      if len(invertedSols) > 1:
        print(f'  inverted integral solutions: ')
        for sol in invertedSols:
          print('    ', sol)
      else:
        print(f'  inverted integral solution: ', invertedSols[0])


  def _setConstants(self, **kwargs):
    # store passed constants dictionary for later reference
    self._constantsDict = kwargs

    # prepare expression object
    if self._probabilityDensityBaseExpr is None:
      self._probabilityDensityBaseExpr = sy.sympify(self._probabilityDensity)
    expr = self._probabilityDensityBaseExpr    

    # substitute constants
    for name, val in kwargs.items():
      expr = expr.subs(name, val)

    # set variables attribute if not set
    self._variables = list(expr.free_symbols)

    # fix variable order if given
    if self._variableOrder:
      sortedVars = []
      for varName in self._variableOrder:
        varNames = [str(v) for v in self._variables]
        if varName in varNames:
          sortedVars.append(self._variables.pop(varNames.index(varName)))
      self._variables = sortedVars + self._variables

    # substitute remaining free symbols with symbols that 
    # have 'real' assumption
    _newVariables = []
    for i, sym in enumerate(self._variables):
      l1, l2 = self._variableDomains.get(str(sym), (-inf, inf))
      realSym = sy.Symbol(str(sym), real=True, 
                          **(dict(nonnegative=True) if l1>=0
                        else dict(nonpositive=True) if l2<=0
                        else {}))
      expr = expr.subs(sym, realSym)
      _newVariables.append(realSym)
    self._variables = _newVariables

    # append variables that appear in domains but not in expression
    varNames = [str(v) for v in self._variables]
    for symName in self._variableDomains.keys():
      if symName not in varNames:
        self._variables.append(sy.Symbol(symName, real=True))

    # save resulting expr in attribute
    self._probabilityDensityExpr = expr


  def _generateAnalyticScalarLambda(self, varI):
    '''
    for lambda for variable number varI integrate over full domain 
    for all var<varI and leave open any var>varI 
    '''
    # prepare symbols and domains
    expr = self._probabilityDensityExpr

    try:
      # start alarm that raises exception in this thread after timeout
      _setAlarm(self._deadline)

      # test whether expression looks positive
      isPositive = False
      try:
        if not bool( expr < 0 ):
          isPositive = True
      except Exception:
        pass
      if not isPositive:
        try:
          if not bool( sy.solve(expr < 0) ):
            isPositive = True
        except Exception:
          pass
      if not isPositive:
        io.warn(f'not sure whether expression for probability density '
                f'"{expr}" always yields positive values; negative '
                f'probabilities will lead to undefined behavior')

      # integrate along domain for i<varI
      for i in range(varI):
        var = self._variables[i]
        l1, l2 = self._variableDomains.get(str(var), (-inf, inf))
        expr = sy.Integral(expr, (var,l1,l2)).doit()

      # integrate and invert for requested var
      var = self._variables[varI]
      l1, l2 = self._variableDomains.get(str(var), (-inf, inf))

      varX = sy.Symbol('x', real=True, **(dict(positive=True) if l1>=0
                                    else dict(negative=True) if l2<=0
                                    else {}))
      varY = sy.Symbol('y', real=True, nonnegative=True)

      # find total and 
      totalIntegral = sy.Integral(expr, (var,l1,l2)).doit()
      partialIntegral = sy.Integral(expr, (var,l1,varX)).doit()
      
      exprYs = sy.solve(sy.Eq(partialIntegral/totalIntegral, varY), varX, 
                        simplify=False)  # do not simplify, this speeds up the solver a lot
      if len(exprYs) == 0:
        raise ValueError(f'expression {partialIntegral/totalIntegral} seems not to be solvable for {varX}')
      lambYs = [sy.lambdify([varY]+self._variables[varI+1:], exprY, 
                            modules=['numpy', 'scipy'])
                                            for exprY in exprYs]

      # attach expressions to lambda for convenience
      for lam in lambYs:
        lam._origExpressions = (expr/totalIntegral, partialIntegral/totalIntegral, exprYs)

    # re-raise special KeyboardInterrupt raised by timer handler as a RuntimeError,
    # re-raise any other KeyboardInterrupts unchanged
    except KeyboardInterrupt as e:
      if str(e) == 'time is up':
        raise RuntimeError('time is up')
      raise

    finally:
      # disable alarm again
      _clearAlarm()

    return lambYs

  
  def _numericalResolution(self, var):
    # set numerical resolution dict
    if not self._numericalResolutions:
      self._numericalResolutions = 5+int(1e6**(1/len(self._variables)))
    if not type(self._numericalResolutions) is dict:
      self._numericalResolutions = {str(v): self._numericalResolutions for v in self._variables}

    # get numerical resolution and ensure to return odd number
    res = int(round(self._numericalResolutions.get(str(var))))
    if res % 2 == 0:
      return res+1
    return res


  def _generateNumericScalarLambda(self, varI):
    # prepare symbols and domains
    expr = self._probabilityDensityExpr

    # make sure free symbols exactly match variable list
    for s in expr.free_symbols:
      if s not in self._variables:
        raise ValueError(f'probabilty density expression {expr} has free '
                         f'symbol {s} which is not in list of '
                         f'variables {self._variables}')

    # prepare param grid for probability density evaluation
    # additional prepare in-between ranges with values sitting between the
    # original range for use with cumsum later 
    variableRanges = []
    variableRangesInBetween = []
    for var in self._variables:
      l1, l2 = self._variableDomains.get(str(var), (-inf, inf))
      if not isfinite(l1) or not isfinite(l2):
        raise ValueError(f'failed to find analytical solution, numerical '
                         f'solution requires finite limits, but found limits '
                         f'[{l1}, {l2}] for variable {var}')
      _range = linspace(l1, l2, self._numericalResolution(var))
      variableRanges.append(_range)
      variableRangesInBetween.append((_range[1:]+_range[:-1])/2)
    variableGrids = meshgrid(*variableRangesInBetween)

    # evaluate expression
    exprLam = sy.lambdify(self._variables, expr)
    gridProbs = exprLam(*variableGrids)

    # if gridProbs is scalar replace it with array of same shape as variable grid
    if not hasattr(gridProbs, 'shape'):
      gridProbs = variableGrids[0] * 0 + gridProbs

    # check whether probabilities are strictly positive
    if (gridProbs < 0).any():
      raise ValueError(f'found negative probability density, '
                       f'expression: {expr}, variable: {self._variables[varI]}')

    # warn if neighboring points in gridsProbs differ by more than threshold
    for dim in range(len(gridProbs.shape)):
      scale = gridProbs.max()-gridProbs.min()
      if scale < 1e-10:
        scale = 1
      r1, r2 = arange(gridProbs.shape[dim])[:-1], arange(gridProbs.shape[dim])[1:]
      diff = take(gridProbs, r1, axis=dim) - take(gridProbs, r2, axis=dim)
      if abs(diff).max()/scale > self._warnIfDiscretizationStepAbove:
        io.warn(f'numerical evaluation of probability density expression '
                f'{self._probabilityDensityExpr} had jumps larger than '
                f'{1e2*self._warnIfDiscretizationStepAbove:.1f}%')

    # numerically integrate (actually just sum because normalization 
    # happens later anyways) along domain for i<varI
    for _ in range(varI):
      gridProbs = gridProbs.sum(axis=-1)

    # integrate (again actually sum) but insert zeros before to properly start from zero
    # using the in-between grid makes the result an accurate estimate for probability density
    # at a given point in the regular (not-in-between) variable grid
    gridProbs = insert(gridProbs, 0, zeros(gridProbs.shape[:-1]), axis=-1)
    gridProbs = cumsum(gridProbs, axis=-1)

    #print('-=--', varI)
    #print(gridProbs/gridProbs[...,-1].max(), variableRanges[varI])

    # make interpolator function that implements numerical inversion of the integral
    def interpolateResult(x, *params,
                          variableRanges=variableRanges,
                          variableRangesInBetween=variableRangesInBetween,
                          gridProbs=gridProbs, varI=varI):

      # loop over all draw params and assemble result array
      # if params list is empty, just run loop once with empty list as _params
      result = []
      for randIter, _params in enumerate(array(params).T if len(params) else [[]]):
        # make sure iterating over _params is possible
        if not hasattr(_params, '__len__'):
          _params = [_params]

        # if finite number of params is given, also iterate over individual x
        # if not params are given, keep shape of x to boost calculation via numpy broadcasting
        if hasattr(x, '__len__') and len(_params):
          _x = x[randIter]
        else:
          _x = x

        # select columns according to conditional variable values
        index = []
        for i, _param in enumerate(_params):
          index.append(argmin(abs(variableRangesInBetween[varI+i+1]-_param)))
        #print(f'{shape(gridProbs)=}, {index=}')
        _gridProbsCol = gridProbs[*index,:]

        # normalize to maximum entry
        _gridProbsCol /= _gridProbsCol[-1]

        # append interpolated values to result
        #print(f'{shape(_gridProbsCol)=}, {shape(variableRanges[varI])=}')
        result.append( interp(_x, _gridProbsCol, variableRanges[varI]) )

      #print(f'{shape(x)=}, {shape(params)=}, {shape(result)=}')

      # transpose results to return in same shape as params were given
      if len(params):
        #print('return array')
        return array(result).T

      # if no params were given return first result only (because there is only one in this case)
      #print('return first')
      return result[0]

    # numerically invert using interpolator
    lambYs = [interpolateResult]
    # attach placeholders instead of expressions to lambda
    for lam in lambYs:
      lam._origExpressions = ('n.a.', 'n.a.', ['n.a.'])

    return lambYs


  def draw(self, N=None, constants=None):
    '''
    Draw random numbers (or vectors) following the distribution represented by this object.

    Arguments:
    ==========
    N : int
      number of random numbers (vectors) to draw. If this argument is not passed, the return
      value will have the same shape as the random number variable implied by the distribution.
      If N is passed, the result will have an additional dimension of size N.

    constants : dict    
      Dictionary of constants to substitute in the distribution expression.
    '''

    # compile variable first if either constants are passed or
    # if it was not yet compiled
    if not hasattr(self, '_transformLambdas') or (constants is not None and constants != self._constantsDict):
      self.compile(**(constants or {}))

    # accept float values for N and limit to min 1
    if N is not None:
      N = max([1, int(round(N))])

    result = []
    for i, transforms in reversed(list(enumerate(self._transformLambdas))):
      #print(f'drawing var {self._variables[i]}...')
      l1, l2 = self._variableDomains.get(str(self._variables[i]), (-inf, inf))

      # roll standard uniform [0,1) rng and transform result, use numpy broadcasting
      # for improved performance
      rand = random.random_sample(**({} if N is None else dict(size=N)))
      #print(f'{shape(result)=}')
      vals = array([transform(rand, *result[::-1]) for transform in transforms])

      # make sure shapes match (only needed for debugging)
      if shape(vals) != (1,1) and shape(vals) != shape([rand]*len(transforms)):
        raise ValueError(f'shape mismatch {shape(vals)=} != {shape([rand]*len(transforms))=}, do '
                         f'all arguments/attributes of this object have intended shapes?')

      # find indices of resulting values that are within bounds
      valid = argwhere(logical_and(l1 <= vals, vals <= l2))

      # make sure each of the N rolls had exactly one valid result
      if any(valid[:,-1] != arange(valid.shape[0])):
        raise ValueError('no/more than one valid value found in domain')

      # append valid results to list
      if N is not None:
        result.append(vals[tuple(valid.T)])
      else:
        result.append(vals[tuple(valid.T)][0])
    
    # reverse result ordering to restore correct variable order
    result = array(result[::-1])

    # return results as dictionary with variable names as keys
    if self._variableOrder is None:
      return {str(k): v for k, v in zip(self._variables, result)}

    # if variable order is specified, return as array with first dimension ordered as given
    # first make sure variableOrdering has 1:1 match with _variables
    _varNames = [str(v) for v in self._variables]
    for v in self._variableOrder:
      if v not in _varNames:
        raise ValueError(f'variable {v} is given in variable ordering, but does not seem to '
                         f'exist in expression {self._probabilityDensityExpr}')
      _varNames.remove(v)
    if len(_varNames):
      raise ValueError(f'variables {_varNames} exist in expression {self._probabilityDensityExpr}'
                       f' but do not exist in {self._variableOrder}; are all constants specified?')

    # construct ordering index and return
    _orderingIndex = [[str(v) for v in self._variables].index(_v) for _v in self._variableOrder]
    return result[_orderingIndex]

  def drawPseudo(self, N, bins=None, overdrawFactor=0.1, overdrawIterations=50, constants=None, plotHistograms=False):
    '''
    Draw pseudo random (or vectors) almost following the distribution represented by this object.
    The histogram if the returned numbers will be close to the expected histogram, outliers are
    removed.

    Arguments:
    ==========
    N : int
      Number of random numbers (vectors) to draw. This must be greater than 1, because a histogram
      is not well defined otherwise.

    bins : int or list
      Optionally pass number of bins to consider

    constants : dict    
      Dictionary of constants to substitute in the distribution expression.

    overdrawFactor : float
      Draw N*overdrawFactor true random numbers and return only the N numbers out of these, that
      represent the expected histogram best.

    overdrawIterations : int
      Redraw new values overdrawIterations times to refine histogram 
    '''
    if N <= 1:
      raise ValueError(f'N must be greater than one in pseudo random mode')
    if overdrawFactor <= 0:
      raise ValueError(f'overdrawFactor must be greater than zero')
    if overdrawIterations <= 1:
      raise ValueError(f'overdrawIterations must be greater than one')
    if not self._variableOrder:
      raise ValueError(f'variableOrder must be passed to constructor to use pseudo random mode.')

    draws = None
    for _ in range(round(int(overdrawIterations))):
      # draw true random samples
      newDraws =  self.draw(N=round(N*overdrawFactor), constants=constants)
      if draws is None:
        # draw (1+factor)*N in the beginning instead of factor*N
        draws = self.draw(N=round(N*(1+overdrawFactor)), constants=constants)
      elif len(draws.shape) == 1:
        # concatenate with old non-nan samples
        draws = concatenate([draws[logical_not(isnan(draws))], newDraws])
      else:
        # concatenate with old non-nan samples
        draws = concatenate([draws[...,logical_not(isnan(draws[0]))], newDraws], axis=-1)

      # calc n-D histogram
      if bins is None:
        bins = int( (overdrawFactor*sqrt(overdrawIterations)*N)**(1/(3*len(self._variableOrder))) )
      hist, edges = histogramdd(draws.T, bins=bins)
      binCenters = [(e[1:]+e[:-1])/2 for e in edges]

      # calc expected histogram
      expr = self._probabilityDensityExpr
      lambdExpr = sy.lambdify(list(reversed(self._variableOrder)), expr)
      expectedHist = lambdExpr(*meshgrid(*reversed(binCenters)))

      # fix shape in case of missing variables in expression
      if not hasattr(expectedHist, 'shape'):
        expectedHist = expectedHist*ones(hist.shape)

      while True:
        # drop values in bins that have more than expected values'
        histDiff = hist/hist.sum() - expectedHist/expectedHist.sum()
        outlierIndex = argwhere(histDiff == histDiff.max())[0]
        outlierEdge1 = [e[i] for e,i in zip(edges, outlierIndex)]
        outlierEdge2 = [e[i+1] for e,i in zip(edges, outlierIndex)]

        #print(outlierEdge1, outlierEdge2)
        isInBin = None
        if len(draws.shape) == 1:
          isInBin = logical_and(outlierEdge1[0] < draws, draws <= outlierEdge2[0])
        else:
          for i, (e1, e2) in enumerate(zip(outlierEdge1, outlierEdge2)):
            _isInBin = logical_and(e1 < draws[i], draws[i] <= e2)
            if isInBin is None:
              isInBin = _isInBin
            else:
              isInBin = logical_and(isInBin, _isInBin)
        drawDeleteIndices = argwhere(isInBin)
        #print(draws[...,drawDeleteIndices[0]])

        # replace random value from that bin with nan and reduce respective histogram count by one,
        # if that bin has no values left, exit loop
        if len(drawDeleteIndices) > 0:
          drawDeleteIndex = drawDeleteIndices[int(random.random()*drawDeleteIndices.shape[0])]
          draws[...,*drawDeleteIndex] = nan
          hist[*outlierIndex] -= 1
          #print(f'hist reduced at {outlierIndex=}, removed draw index [{drawDeleteIndex=}]')
        else:
          draws = draws[...,logical_not(isnan(draws if len(draws.shape)==1 else draws[0]))]
          draws = draws[...,-int(N):]
          #print('end prematurely')
          break

        # end loop if only N values left
        if len(draws.shape) == 1:
          if sum(logical_not(isnan(draws))) <= N:
            #print('end because need more samples')
            break
        else:
          if sum(logical_not(isnan(draws[-1]))) <= N:
            #print('end because need more samples')
            break

      if plotHistograms:
        histDiff = hist/hist.sum() - expectedHist/expectedHist.sum()
        import matplotlib.pyplot as plt 
        plt.figure(figsize=(5,2))
        plt.plot(hist/hist.sum())
        plt.plot(expectedHist/expectedHist.sum())
        plt.plot(histDiff)

    # clean out nans and return, make sure length is right and return
    result = draws[...,logical_not(isnan(draws if len(draws.shape)==1 else draws[0]))]
    if draws.shape[-1] / result.shape[-1] > 5:
      warnings.warn('pseudo random generation was not very successful, maybe bins '
                    'or overdraw parameters have to be tweaked...')
    return result[...,-int(round(N)):]


  def findGrid(self, N, startFrom=None, constants=None):
    '''
    Return values whose density follows the given probability density as close as possible.
    '''
    # compile variable first if either constants are passed or
    # if it was not yet compiled
    if not hasattr(self, '_transformLambdas') or (constants is not None and constants != self._constantsDict):
      self.compile(**(constants or {}))

    # perform 1D grid generation
    if len(self._variables) == 1:
      # prepare ranges and parameters      
      var = self._variables[0]
      l1, l2 = self._variableDomains.get(str(var), (-inf, inf))
      if not isfinite(l1) or not isfinite(l2):
        raise ValueError(f'variable domains must be finite for grid generation')
      varRange = linspace(l1, l2, self._numericalResolution(var))

      # calc density
      expr = self._probabilityDensityExpr
      lambdExpr = sy.lambdify(var, expr)
      density = lambdExpr(varRange)

      # fix shape in case of missing variables in expression
      if not hasattr(density, 'shape'):
        density = density*ones(varRange.shape)

      # find startFrom parameter if not given explicitly
      if startFrom is None:
        startFrom = varRange[argmax(density)]
      
      # call algorithm from points_by_density module
      result = points_by_density.generatePointsWithGivenDensity1D(
                              density=(varRange, density),
                              N=N,
                              startFrom=startFrom)

      # clip result to range and return
      return result[logical_and(min(varRange) <= result, result <= max(varRange))]
    else:
      raise RuntimeError('grid generation is not implemented for variable count greater than 1')



class ScalarRandomVariable(VectorRandomVariable):
  '''
  Scalar valued random variable. 
  '''
  def __init__(self, probabilityDensity, variableDomain, variable=None, numericalResolution=None, **kwargs):
    self._desiredVariable = variable
    if variable is None:
      variable = str(list(sy.sympify(probabilityDensity).free_symbols)[0])
    super().__init__(probabilityDensity, 
                     variableDomains={variable: variableDomain},
                     numericalResolutions={} if numericalResolution is None else {variable: numericalResolution},
                     variableOrder=[variable,],
                     **kwargs)

  def compile(self, **kwargs):
    # subfunction that raises human readable exceptions if conditions for scalar random variable are not fulfilled 
    def _checkScalarity():
      freeSymbols = sy.sympify(self._probabilityDensityExpr).free_symbols
      if ( len(freeSymbols) 
            and self._desiredVariable is not None
            and self._desiredVariable not in [str(s) for s in freeSymbols] ):
        raise ValueError(f'specified variable "{self._desiredVariable}" does not seem to appear '
                         f'in expression "{self._probabilityDensityExpr}"')

      if len(self._variables) > 1:
        raise ValueError(f'expression "{self._probabilityDensityExpr}" seems to have more than '
                        f'one free variable after substituting constants; did you pass all constants '
                        f'to .compile() or .draw()?')

    # run vector random variable's compile and reraise exceptions caused by non-scalarity as 
    # useful human readable exceptions
    try:
      super().compile(**kwargs)
    except ValueError as e:
      if 'requires finite limits' in str(e):
        _checkScalarity()
      raise
    _checkScalarity()

  def draw(self, N=None, **kwargs):
    return super().draw(N=N, **kwargs)[0]



__license__ = 'LGPL-3.0-or-later'
__copyright__ = 'Copyright 2024  W. Braun (epiray GmbH)'
__authors__ = 'P. Bredol'
__url__ = 'https://github.com/zaphB/freecad.optics_design_workbench'


from numpy import *
import pickle
import time
import os

from ..detect_pyside import *
from . import common

_PROGRESS_WINDOW = None


def secondsToYMDhms(secs):
  res = []
  for mult in [365*24*60*60, 30*24*60*60, 24*60*60, 60*60, 60, 1]:
    res.append(int(secs/mult))
    secs -= res[-1]*mult
  return res

def secondsToStr(secs, length=2):
  res = []
  split = secondsToYMDhms(secs)
  for num, label in zip(split,
                        ['Y', 'M', 'd', 'h', 'm', 's']):
    if res or num or label=='s':
      res.append(str(num) + label)
  return ' '.join(res[:length])

def scaleSuff(x):
  if x >= 2e9:
    return 1e-9, 'G'
  if x >= 2e6:
    return 1e-6, 'M'
  if x >= 2e3:
    return 1e-3, 'k'
  return 1, ''

class QLabeledProgress(QProgressBar):
  def __init__(self, label, maximum=None):
    super().__init__()
    self.setRange(0,1)
    self.maximum = maximum or inf
    self.label = label
    self.t0 = time.time()
    self.remainingSeconds = None
    self.lastVal = None

  def setValue(self, val):
    self.lastVal = val
    scale, suff = scaleSuff(val)

    # decide out how many digits to show
    digits = 2
    if val*scale>=10:
      digits = 1
    if val*scale>=1e2 or scale==1:
      digits = 0

    # show progress including target value
    if isfinite(self.maximum):
      if time.time()-self.t0 > 5 and val > 0:
        self.remainingSeconds = (time.time()-self.t0)/val * max([self.maximum-val, 0])
      self.setRange(0, max([val+1e-2, self.maximum]))
      mscale, msuff = scaleSuff(self.maximum)
      # decide how many digits to show for maximum
      mdigits = 1
      if self.maximum*mscale>=10 or mscale==1:
        mdigits = 0
      self.setFormat(f'{self.label}: {val*scale:.{digits}f}{suff}'
                     f' / {self.maximum*mscale:.{mdigits}f}{msuff}')
      super().setValue(val)

    # or just show number and never increment bar
    else:
      self.setFormat(f'{self.label}: {val*scale:.{digits}f}{suff}')
      super().setValue(0)


class ProgressWindow(QWidget):
  '''
  ProgressWindows are linked to a SimulationResults instance and constantly 
  consume the progress files being dumped into the progress directory. The contents
  of the consumed files will then be displayed in the window.
  '''
  def __init__(self):
    # setup window contents
    super().__init__()
    self.setWindowTitle('FreeCAD Optics Design Workbench simulation progress')
    layout = QVBoxLayout()
    self.label = QLabel()
    layout.addWidget(self.label)
    self.lastUpdatedLabel = 0

    self.iterations = QLabeledProgress('iterations')
    layout.addWidget(self.iterations)

    self.raysTraced = QLabeledProgress('rays')
    layout.addWidget(self.raysTraced)

    self.hitsRecorded = QLabeledProgress('recorded hits')
    layout.addWidget(self.hitsRecorded)

    self.setLayout(layout)

    # setup timer to consume files
    self.timer = QTimer()
    self.timer.timeout.connect(self.onTimer)
    self.timer.start(50)

  def setStore(self, store):
    self.store = store
    if any([isfinite(m) for m in (store.endAfterIterations, store.endAfterRays, store.endAfterHits)]):
      self.label.setText('simulation progress (approx. ..... remain)')
    else:
      self.label.setText('simulation progress (running infinitely)')
    self.iterations.maximum = store.endAfterIterations
    self.raysTraced.maximum = store.endAfterRays
    self.hitsRecorded.maximum = store.endAfterHits
    self.iterations.t0 = time.time()
    self.iterations.remainingSeconds = None
    self.raysTraced.t0 = time.time()
    self.raysTraced.remainingSeconds = None
    self.hitsRecorded.t0 = time.time()
    self.hitsRecorded.remainingSeconds = None
    self.timer.start(50)

  def onTimer(self):
    progress = self.store.getProgress()
    remainingSeconds = inf

    # update all progress bars
    for bar, key in [(self.iterations,   'totalIterations'),
                     (self.raysTraced,   'totalTracedRays'),
                     (self.hitsRecorded, 'totalRecordedHits'),
                    ]:
      # calculate sum of all workers
      bar.setValue(progress.get(key, 0))

      # if bar offers remaining seconds field show it
      if bar.remainingSeconds is not None:
        remainingSeconds = min([remainingSeconds, bar.remainingSeconds])

    if remainingSeconds < 1e9 and time.time()-self.lastUpdatedLabel > .5:
      self.label.setText(f'simulation progress (approx. {secondsToStr(remainingSeconds)} remain)')
      self.lastUpdatedLabel = time.time()

    # stop timer if simulation is not running
    if not self.store.isSimulationRunning():
      status = 'done' if self.store.simulationEndedGracefully() else 'canceled'
      self.label.setText(f'simulation progress ({status})')
      self.timer.stop()


def showProgressWindow(store):
  global _PROGRESS_WINDOW
  
  # ignore call if gui windows are set to be hidden
  if common.HIDE_GUI:
    return

  # only create window if inside a QtApp
  if QApplication.instance() is not None:
    if _PROGRESS_WINDOW is None:
      _PROGRESS_WINDOW = ProgressWindow()
    _PROGRESS_WINDOW.setStore(store)
    _PROGRESS_WINDOW.show()



__license__ = 'LGPL-3.0-or-later'
__copyright__ = 'Copyright 2024  W. Braun (epiray GmbH)'
__authors__ = 'P. Bredol'
__url__ = 'https://github.com/zaphB/freecad.optics_design_workbench'


try:
  import FreeCADGui as Gui
  import FreeCAD as App
except ImportError:
  pass

import time

from .. import simulation

_LAST_PROCESS_EVENTS_CALL = time.time()
_MIN_UPDATE_INTERVAL = 1e-2

class SimulationEnded(RuntimeError):
  pass

def keepGuiResponsive(raiseIfSimulationDone=False):
  from ..detect_pyside import QApplication  
  global _LAST_PROCESS_EVENTS_CALL
  if time.time()-_LAST_PROCESS_EVENTS_CALL > _MIN_UPDATE_INTERVAL:
    _LAST_PROCESS_EVENTS_CALL = time.time()

    if QApplication.instance():
      # process Qt events
      QApplication.processEvents()
      Gui.updateGui()
      QApplication.processEvents()

    # check whether simulation was canceled and raise SimulationEnded if so
    if raiseIfSimulationDone and (simulation.isCanceled() or simulation.isFinished()):
      raise SimulationEnded()
      
def keepGuiResponsiveAndRaiseIfSimulationDone():
  keepGuiResponsive(raiseIfSimulationDone=True)



__license__ = 'LGPL-3.0-or-later'
__copyright__ = 'Copyright 2024  W. Braun (epiray GmbH)'
__authors__ = 'P. Bredol'
__url__ = 'https://github.com/zaphB/freecad.optics_design_workbench'


import FreeCADGui as Gui
import FreeCAD as App

from . import freecad_elements

class OpticsDesignWorkbench(Gui.Workbench):
  MenuText = 'Optics Design'
  ToolTip = 'Ray tracing Monte-Carlo simulation for optics design and optimization'
  Icon = freecad_elements.find.iconpath('workbench')
  toolbox = []

  def Initialize(self):
      '''This function is executed when FreeCAD starts'''
      # import here all the needed files that create your FreeCAD commands
      self.toolbox = [
        # sources
        'Add point source', 
        'Add replay source', 

        # optical elements
        'Make mirror', 
        'Make lens', 
        'Make grating', 

        # detectors
        'Make absorber', 
        'Make detector', 

        # place settings node
        'Insert settings',

        # place/start/stop simulation in the various modes
        'Clear all rays',
        'Place ray fans',
        'Single pseudo random',
        'Single true random',
        'Continuous pseudo random',
        'Continuous true random',
        'Stop continuous',
      ]
      
      self.appendToolbar(self.__class__.MenuText, self.toolbox)
      self.appendMenu(self.__class__.MenuText, self.toolbox)

  def Activated(self):
      '''This function is executed when the workbench is activated'''
      return

  def Deactivated(self):
      '''This function is executed when the workbench is deactivated'''
      return

  def ContextMenu(self, recipient):
      '''This is executed whenever the user right-clicks on screen'''
      self.appendContextMenu(self.__class__.MenuText, self.toolbox)

  def GetClassName(self):
      # this function is mandatory if this is a full python workbench
      return 'Gui::PythonWorkbench'

Gui.addWorkbench(OpticsDesignWorkbench())
freecad_elements.loadAll()



'''
Definitions for behavior/objects/commands living in the FreeCAD projects. The ray-tracing code is in here, because it is inherently connected to the FreeCAD geometry engine.
'''

__license__ = 'LGPL-3.0-or-later'
__copyright__ = 'Copyright 2024  W. Braun (epiray GmbH)'
__authors__ = 'P. Bredol'
__url__ = 'https://github.com/zaphB/freecad.optics_design_workbench'

from .point_source import *
from .replay_source import *
from .optical_group import *
from .simulation_settings import *
from .simulation_actions import *
from .common import *
from . import find

def loadAll():
  '''
  Load all FreeCAD components defined in the submodules and add them to the interface.
  '''
  loadPointSource()
  loadReplaySource()
  loadGroups()
  loadSimulationSettings()
  loadSimulationActions()



'''

'''

__license__ = 'LGPL-3.0-or-later'
__copyright__ = 'Copyright 2024  W. Braun (epiray GmbH)'
__authors__ = 'P. Bredol'
__url__ = 'https://github.com/zaphB/freecad.optics_design_workbench'

from numpy import *
from matplotlib.pyplot import *
import seaborn as sns
import pandas as pd
import functools

from .. import io
from . import histogram

nx, ny, nz = array([1,0,0]), array([0,1,0]), array([0,0,1])

class Hits:
  def __init__(self, hits):
    self.hits = hits

  def __iter__(self):
    return self.hits.keys()

  def __len__(self):
    return len(self.points())

  def items(self):
    return self.hits.items()

  def keys(self):
    return self.hits.keys()

  def values(self):
    return self.hits.values()

  def points(self):
    if 'points' in self.hits:
      return self.hits['points']
    return array([])

  def directions(self):
    if 'directions' in self.hits:
      return self.hits['directions']
    return array([])

  def isEntering(self):
    if 'isEntering' in self.hits:
      return self.hits['isEntering']
    return array([])

  # ====================================================
  # POINT CLOUD MATH

  def planeProject3dPoints(self, points=None, planeNormal=None, 
                           xInPlaneVec=None, returnZ=False):
    '''
    Turn a 3D point cloud of shape (N,3) into a 2D point cloud (N,2).

    Parameters:
      points : (N,3) array (optional)
        Point cloud to project, defaults to self.points().

      planeNormal : (3,) array (optional)
        Normal vector of the projection plane, defaults to result
        of detectPlaneNormal(points).

      xInPlaneVec : (3,) array (optional)
        Vector pointing into the desired X-coordinate direction of
        the coordinate system after projection. Defaults to the 
        numerically best suited vector among (1,0,0), (0,1,0) and (0,0,1).
    '''
    # detect plane normal if none is given explicitly
    if points is None:
      points = self.points()
    if planeNormal is None or xInPlaneVec is None:
      planeNormal, xInPlaneVec = self.detectPlaneNormal(planeNormal=planeNormal, 
                                                        xInPlaneVec=xInPlaneVec)

    projX = xInPlaneVec
    X = dot(points, projX/linalg.norm(projX)) 
    projY = cross(planeNormal, xInPlaneVec)
    Y = dot(points, projY/linalg.norm(projY))
    if returnZ:
      Z = dot(points, planeNormal/linalg.norm(planeNormal))
      return array([X, Y, Z]).T
    return array([X, Y]).T

  def detectPlaneNormal(self, points=None, directions=None, 
                        planeNormal=None, xInPlaneVec=None,
                        maxPointCountConsidered=300, angleTol=1e-9):
    '''
    Find the best possible plane normal vector to project the 3D point cloud
    points to a 2D point cloud spanning a maximal area.
    '''
    if points is None:
      points = self.points()
    if directions is None:
      directions = self.directions()
      isEntering = self.isEntering()
      # remove directions of exiting rays when more rays entered than exited
      # -> heuristic to improve planeNormal sign auto-detect for transparent detectors
      if sum(isEntering==0) < .51*len(isEntering):
        directions = directions[isEntering!=0]
    checkPoints = points[::1+int(points.shape[0]/maxPointCountConsidered)]
    checkDirections = directions[::1+int(directions.shape[0]/maxPointCountConsidered)]

    if planeNormal is None:
      # cover only half the unit sphere because every plane has two normal
      # vectors (sign is ambiguous)
      phis = linspace(0, pi, 30)
      dphi = phis[1]-phis[0]
      thetas = linspace(-pi/2, pi/2, 30)
      dtheta = thetas[1]-thetas[0]
      while True:
        # check all phi theta candidates and calculate expected span for them
        phiThetas, zSpans = [], []
        for phi, theta in zip(*[g.flatten() for g in meshgrid(phis, thetas)]):
          normal = array([cos(phi)*sin(theta), sin(phi)*sin(theta), cos(theta)])
          p = dot(checkPoints, normal)
          phiThetas.append((phi, theta))
          zSpans.append( p.max()-p.min() )

        # select phi thata with the best span so far and update grid
        phiOpt, thetaOpt = phiThetas[argmin(zSpans)]
        phis = linspace(phiOpt-1.1*dphi, phiOpt+1.1*dphi, 10)
        dphi = phis[1]-phis[0]
        thetas = linspace(thetaOpt-1.1*dtheta, thetaOpt+1.1*dtheta, 10)
        dtheta = thetas[1]-thetas[0]

        # quit loop if angle diff is smaller than tolerance
        if dphi < angleTol and dtheta < angleTol:
          break

      # calc best planeNormal
      planeNormal = array([cos(phiOpt)*sin(thetaOpt), sin(phiOpt)*sin(thetaOpt), cos(thetaOpt)])

    # choose planeNormal sign to be parallel with directions as well as possible
    projectedDirs = dot(checkDirections, planeNormal)
    # 90% of directions are same direction as planeNormal -> flip plane normal
    if quantile(projectedDirs, 0.1) > 0:
      planeNormal = -planeNormal
    # 90% of directions are opposite direction of planeNormal -> all good
    elif quantile(projectedDirs, 0.9) < 0:
      pass
    # something in between: follow best guess and warn user
    else:
      if quantile(projectedDirs, 0.5) < 0:
        planeNormal = -planeNormal
      io.warn(f'unsure of result when trying to auto-detect sign of plane normal, '
              f'avoid relying on the sign of the planeNormal')

    # select projection vectors by cross products of normal with 
    # nx, ny and nz with highest norm to ensure numeric stability
    # if planeNormal is almost parallel to one of nx, ny or nz.
    candidates = [nx, ny, nz]
    if xInPlaneVec is not None:
      candidates = [xInPlaneVec]
    projY = sorted([cross(planeNormal, n) for n in candidates], 
                   key=lambda x: -linalg.norm(x))[0]
    xInPlaneVec = cross(planeNormal, projY)
    # prefer positive coefficients in xInPlaneVec
    if sum(xInPlaneVec) < 0:
      xInPlaneVec = -xInPlaneVec

    # set plane normal to optimal normal found so far
    return planeNormal, xInPlaneVec

  def histogram(self, planeNormal=None, xInPlaneVec=None, key='points', **kwargs):
    '''
    Return the histogram of a 3D-point-cloud projected to a planeNormal. If planeNormal
    is not passed, a normal orthogonal to the two largest extents of the point could
    is chosen.
    Returns three rays: histogram, X and Y axes values. X and Y are either the bin edges
    (XY='edges') or the bin centers (XY='center')
    '''
    points = self.hits[key]

    # if project vector is not given, set it to the direction of minimum span
    if planeNormal is None or xInPlaneVec is None:
      planeNormal, xInPlaneVec = self.detectPlaneNormal(planeNormal, xInPlaneVec)

    # perform projection of full dataset
    projPoints = self.planeProject3dPoints(points, planeNormal=planeNormal, xInPlaneVec=xInPlaneVec)
    X, Y = projPoints.T
    return histogram.Histogram(X, Y, planeNormal=planeNormal, xInPlaneVec=xInPlaneVec, **kwargs)


  def plot(self, hueKey=None, hueLabel=None, planeNormal=None, xInPlaneVec=None, 
           plotKey='points', **kwargs):
    # just return if no hits exist
    if plotKey not in self.hits.keys():
      return

    if planeNormal is None or xInPlaneVec is None:
      planeNormal, xInPlaneVec = self.detectPlaneNormal(points=self.hits[plotKey], 
                                                        planeNormal=planeNormal, 
                                                        xInPlaneVec=xInPlaneVec)
    X, Y = self.planeProject3dPoints(self.hits[plotKey], planeNormal=planeNormal).T
    data = {'projected $x$':X, 'projected $y$':Y}
    if hueKey is not None:
      if hueLabel is None:
        hueLabel = hueKey
      data[hueLabel] = self.hits[hueKey]
    sns.scatterplot(pd.DataFrame(data),
                    x='projected $x$', y='projected $y$', 
                    **(dict(hue=hueLabel, palette='hls') if hueLabel else {}),
                    **kwargs)
    nx, ny, nz = planeNormal
    px, py, pz = xInPlaneVec
    title(f'plane normal = [{nx:.2f}, {ny:.2f}, {nz:.2f}],\n'
          f'projected $x$ = [{px:.2f}, {py:.2f}, {pz:.2f}]', fontsize=10) 
    gca().axis('equal')
    gca().set_aspect('equal')
    tight_layout()

  # ====================================================
  # FAN MATH

  def supportsFanMath(self):
    return all([k in self.hits.keys() for k in 'rayIndex fanIndex totalRaysInFan'.split()])

  def _raiseIfNotFanMath(self):
    if not self.supportsFanMath():
      raise ValueError(f'keys rayIndex, fanIndex and totalRaysInFan must exist in hits dictionary, '
                f'make sure you '
                f'simulated in fan mode and enabled storing the respective metadata keys '
                f'in the active SimulationSettings')

  def raysPerFan(self):
    self._raiseIfNotFanMath()
    return self.hits['totalRaysInFan'][0]

  def allRayIndices(self, fanI=None):
    rI, fI = self.hits['rayIndex'], self.hits['fanIndex']
    if fanI is not None:
      return array(sorted(list(set(rI[fI==fanI]))))
    return array(sorted(list(set(rI))))

  def fanCount(self):
    self._raiseIfNotFanMath()
    return len(set(self.hits['fanIndex']))

  @functools.cache
  def _calcFanDensityEtc(self, pCenter=None, **kwargs):
    self._raiseIfNotFanMath()

    rI, fI, p, trf = self.hits['rayIndex'], self.hits['fanIndex'], self.hits['points'], self.hits['totalRaysInFan']
    pXY = self.planeProject3dPoints(p, **kwargs)

    # loop through fans/rays and calculate distance, curvature etc.
    centerDists, neighborDists = [], []
    curvs, missingRays, skippedRays = [], 0, 0
    for fanI in sorted(list(set(fI))):
      prevPrevPos = None
      prevPos = None
      rayIs = sorted(list(set(rI[fI==fanI])))

      # find most central ray
      if pCenter is None:
        pCenter = self.fanCenter()
      pCenter = array(pCenter)

      # increment missing rays
      missingRays += mean(trf[fI==fanI])-len(rayIs)
      skippedRays += sum(array(rayIs[1:])-array(rayIs[:-1]) - 1)

      # find most likely vectors of positive and negative ray index direction
      _posIndexVectors = pXY[logical_and(fI==fanI, rI>0)]-pCenter
      posIndexDirection = None
      if len(_posIndexVectors):
        _norms = sqrt(sum(_posIndexVectors**2, axis=0))
        _norms[_norms==0] = 1
        posIndexDirection = mean(_posIndexVectors/_norms, axis=0)
      _negIndexVectors = pXY[logical_and(fI==fanI, rI<0)]-pCenter
      negIndexDirection = None
      if len(_negIndexVectors):
        _norms = sqrt(sum(_negIndexVectors**2, axis=0))
        _norms[_norms==0] = 1
        negIndexDirection = mean(_negIndexVectors/_norms, axis=0)

      # fill missing directions with respective other of opposite sign
      if posIndexDirection is None and negIndexDirection is None:
        posIndexDirection = array([1,0])
        negIndexDirection = array([-1,0])
      elif posIndexDirection is None:
        posIndexDirection = -negIndexDirection
      elif negIndexDirection is None:
        negIndexDirection = -posIndexDirection

      # loop through ray-trios to calculate neighbor dists and curvs
      for i1, i0, i2 in zip(rayIs+[None,None], [None]+rayIs+[None], [None,None]+rayIs):        

        # calc points and distances
        p0, p1, p2 = [None if i is None else mean(pXY[logical_and(fI==fanI, rI==i)], axis=0) 
                                                                      for i in (i0, i1, i2)]
        d1, d2, dCenter = [None if _p is None or __p is None else  sqrt(sum((_p-__p)**2)) 
                                              for _p,__p in ([p0,p1], [p0,p2], [p0,pCenter])]

        # add neighbor and center dist entries
        if d1 is not None:
          neighborDists.append( [ fanI, (i0+i1)/2, d1 ] )
        if i0 is not None:
          # decide about dCenterSign
          signP, signN = dot(p0-pCenter, posIndexDirection), dot(p0-pCenter, negIndexDirection)
          if signP > 0 and signN < 0:
            dCenterSign = +1
          elif signP < 0 and signN > 0:
            dCenterSign = -1
          else:
            if signN != 0 and signP != 0:
              io.warn(f'unsure about center distance value signs, the fan-hit pattern is probably '
                      f'very asymmetric ({dot(p0-pCenter, posIndexDirection)=}, {dot(p0-pCenter, negIndexDirection)=})')
            dCenterSign = sign( signP - signN )

          # add center dist entry
          centerDists.append( [ fanI, i0, dCenter*dCenterSign ] )

        # calc distance of p0 to line through p1,p2
        if i0 is not None and i1 is not None and i2 is not None:
          (x0, y0), (x1, y1), (x2, y2) = (p0, p1, p2)
          curvs.append( [ fanI, i0, abs( (y2-y1)*x0 - (x2-x1)*y0 + x2*y1 - y2*x1 ) / sqrt((y2-y1)**2 + (x2-x1)**2) ] )

    # return as dict
    return dict(centerDists=array(centerDists), neighborDists=array(neighborDists), 
                curvs=array(curvs), missingRays=missingRays, skippedRays=skippedRays,
                rI=rI, fI=fI, pXY=pXY, trf=trf)

  def fanMissingRays(self):
    return self._calcFanDensityEtc()['missingRays']

  def fanSkippedRays(self):
    return self._calcFanDensityEtc()['skippedRays']

  def fanCenterDists(self, pCenter=None):
    return self._calcFanDensityEtc(pCenter=(None if pCenter is None else tuple(pCenter)))['centerDists'].T

  def fanNeighborDists(self):
    return self._calcFanDensityEtc()['neighborDists'].T

  def fanCenter(self, **kwargs):
    self._raiseIfNotFanMath()
    rI, fI, p = self.hits['rayIndex'], self.hits['fanIndex'], self.hits['points']
    pXY = self.planeProject3dPoints(p, **kwargs)

    centers = []
    for fanI in set(fI):
      # zero ray index exists: use zero ray position as center
      if 0 in rI[fI==fanI]:
        centers.extend( pXY[(fI==fanI) & (rI==0)] )

      # no zero ray index exists: use average of +1 and -1 as center 
      elif +1 in rI[fI==fanI] and -1 in rI[fI==fanI]:
        centers.extend(( pXY[(fI==fanI) & (rI==+1)] + pXY[(fI==fanI) & (rI==-1)] )/2)

    # return average center, or nan/nan if no fan center was found
    if len(centers):
      return mean(centers, axis=0)
    return array([nan,nan])

  def fanCurvs(self):
    return self._calcFanDensityEtc()['curvs'].T

  @functools.cache
  def _fanPowerDensityEtc(self, pCenter=None):
    if pCenter is None:
      pCenter = self.fanCenter()

    # get fan neighbor dists and distances from center
    nfI, nrI, ndist = self.fanNeighborDists()
    cfI, crI, cdist = self.fanCenterDists(pCenter=pCenter)
    
    # construct per fan estimated power density
    fanDensities = {}
    causticIntensity = 0
    for fanI in sorted(list(set(nfI))):
      fanDensities[fanI] = []
      for interRayI in sorted(nrI[fanI==nfI]):
        # find two center dists of rays around interRayI (.6 to ensure that neighbors for 
        # interRayI==0 are correctly found as +1 and -1, not 0 and 0 )
        cr1, cr2 = int(round(interRayI-.6)), int(round(interRayI+.6))
        cdist1 = mean( cdist[(fanI==cfI) & (crI==cr1)] )
        cdist2 = mean( cdist[(fanI==cfI) & (crI==cr2)] )

        # find neighbor ray dist
        estimatedPower = 1/mean( ndist[(fanI==nfI) & (nrI==interRayI)] )

        # if cdist is not monotonically increasing with index -> increment 
        # caustic intensity counter:
        if cdist2 < cdist1:
          causticIntensity += abs(cdist1-cdist2)

        # if rays have the expected order, add estimated power and its position to result list:
        else:
          fanDensities[fanI].append([ mean([cdist1, cdist2]), estimatedPower ])

    # construct lambdas to interpolate found densities as a function of 
    # distance from center
    fanDensityFuncs = [(i, lambda pos, _d=array(d).T: interp(pos, *_d, left=0, right=0)) for i, d in fanDensities.items()]

    # return all results as dictionary
    return dict(fanDensities=fanDensities, fanDensityFuncs=fanDensityFuncs,
                causticIntensity=causticIntensity, pCenter=pCenter)
    
  def fanEstimatedPowerDensities(self, pCenter=None):
    return [(i, array(d).T) for i, d in self._fanPowerDensityEtc(pCenter=(None if pCenter is None else tuple(pCenter)))['fanDensities'].items()]
  
  def fanEstimatedPowerDensityFuncs(self, pCenter=None):
    return self._fanPowerDensityEtc(pCenter=(None if pCenter is None else tuple(pCenter)))['fanDensityFuncs']

  def fanEstimatedCausticIntensity(self, pCenter=None):
    return self._fanPowerDensityEtc(pCenter=(None if pCenter is None else tuple(pCenter)))['causticIntensity']



__license__ = 'LGPL-3.0-or-later'
__copyright__ = 'Copyright 2024  W. Braun (epiray GmbH)'
__authors__ = 'P. Bredol'
__url__ = 'https://github.com/zaphB/freecad.optics_design_workbench'


import functools
import subprocess
import sys

from . import io

@functools.cache
def detectQtMajorVersion():
  if 'freecad' in sys.executable.lower():
    r = subprocess.run(['ldd', sys.executable], capture_output=True,
                       text=True, check=False)
    if 'qt5' in r.stdout.lower():
      return 5
    if 'qt6' in r.stdout.lower():
      return 6
  return None


# first try to import pyside without version, this is FreeCAD
# specific and is equivalent to the correct version of pyside
try:
  from PySide.QtCore import *
  from PySide.QtWidgets import *
  #io.verb('used module "PySide"')

except ImportError:
  # second attempt: detect Qt version then decide which pyside 
  # to import
  if detectQtMajorVersion() == 5:
    from PySide2.QtCore import *
    from PySide2.QtWidgets import *
    #io.verb('detected Qt version 5, used module "PySide2"')

  elif detectQtMajorVersion() == 6:
    from PySide6.QtCore import *
    from PySide6.QtWidgets import *
    #io.verb('detected Qt version 6, used module "PySide6"')

  # third attempt: try to import pyside6 first and try older 
  # versions on import error
  else:
    try:
      from PySide6.QtCore import *
      from PySide6.QtWidgets import *
      #io.verb('failed to detect Qt version, used module "PySide6"')
    except ImportError:
      try:
        from PySide2.QtCore import *
        from PySide2.QtWidgets import *
        #io.verb('failed to detect Qt version, used module "PySide2"')
      except ImportError:
        pass



'''

'''

__license__ = 'LGPL-3.0-or-later'
__copyright__ = 'Copyright 2024  W. Braun (epiray GmbH)'
__authors__ = 'P. Bredol'
__url__ = 'https://github.com/zaphB/freecad.optics_design_workbench'


from numpy import *
import os
import subprocess
import time
import signal
import threading
import queue
import random
import functools
import glob
import pickle
import shutil
import traceback
from atomicwrites import atomic_write

from .. import io
from ..simulation import processes
from . import progress
from . import hits
from . import parameter_sweeper

_PRINT_FREECAD_COMMUNICATION = False
_PRINT_SETTER_AND_CALL_LINES = False

# signal handler that kills all running freecad processes
# in case this process is killed
def handler(signum, frame):
  io.info(f'signal handler called with signal {signum}')
  try:
    for doc in _ALL_DOCUMENTS:
      doc.close()
  except Exception:
    io.warn(f'exception raised in signal handler:\n\n'+traceback.format_exc())
  finally:
    exit(signum)

# register signal handlers
for sig in (signal.SIGHUP, signal.SIGTERM):
  signal.signal(sig, handler)

# list that holds all ever opened documents
_ALL_DOCUMENTS = []

# default path to freecad executable
_DEFAULT_FREECAD_EXECUTABLE = 'FreeCAD'


def setDefaultFreecadExecutable(path):
  global _DEFAULT_FREECAD_EXECUTABLE

  # check if path points to executable file if path looks like a filesystem path
  if '/' in path:
    # make sure to expand user
    path = os.path.expanduser(path)

    # check if path exists
    if not os.path.exists(path):
      raise ValueError(f'path {path} does not exist in filesystem')

    # check if regular file (if not on PATH)
    if not os.path.isfile(path):
      raise ValueError(f'path {path} is not looking like a file, did '
                      f'you specify the full path to the FreeCAD executable '
                      f'or AppImage?')

    # check if path is executable (if not on PATH)
    if not os.access(path, os.X_OK):
      raise ValueError(f'path {path} is missing executable rights, did '
                      f'you specify the full path to the FreeCAD executable '
                      f'or AppImage?')

    # replace path with absolute path
    path = os.path.abspath(path)

  else:
    try:
      p = subprocess.run(['whereis', path], capture_output=True, text=True)
    except Exception as e:
      io.warn(f'failed to check whether path exists ({e.__class__.__name__}: {str(e)})')
    else: 
      if not len(''.join(p.stdout.split(':')[1:]).strip()):
        raise ValueError(f'path {path} is not found by whereis')

  # store absolute path in global var
  _DEFAULT_FREECAD_EXECUTABLE = path


def freecadVersion():
  p = subprocess.Popen([_DEFAULT_FREECAD_EXECUTABLE, '-c'],
                        stdout=subprocess.PIPE,
                        stderr=subprocess.DEVNULL,
                        stdin=subprocess.PIPE, 
                        text=True, bufsize=-1)
  try:
    time.sleep(.3)
    p.stdin.write('import FreeCAD; print("version follows"); print(FreeCAD.Version());\n\n')
    p.stdin.flush()
    while True:
      l = p.stdout.readline()
      if 'version follows' in l:
        l = p.stdout.readline()
        return '.'.join([e.strip()[1:-1] for e in l[1:-1].split(',')[:3]])
  except Exception:
    return None
  finally:
    p.stdin.write('\n\nexit();\n\n'*5)
    p.stdin.flush()
    time.sleep(.1)
    if p.poll() is None:
      p.kill()


# create aliases for FreeCAD types that just return simple numpy arrays
# to allow converting Vectors from FreeCAD side automatically
def Vector(*a):
  return array(a)
def Matrix(*a):
  return array(a)



class FreecadExpression:
  def __init__(self, exprString):
    self._exprString = exprString

  def _freecadShellRepr(self):
    return self._exprString

  def __repr__(self):
    return self._exprString


class FreecadPropertyDict:
  def __init__(self, d):
    self._d = d

  def __str__(self):
    return repr(self)

  def __repr__(self):
    return f'<FreecadPropertyDict {repr(self._d)} >'

  def __getattr__(self, k):
    return getattr(self._d, k)

  def __getitem__(self, k):
    return self._d[k]

  def __setitem__(self, k, v):
    self._d[k].set(v)


class FreecadConstraintDict(FreecadPropertyDict):
  def __init__(self, constraintDict, indexDict, sketch):
    self._d = constraintDict
    self._indexDict = indexDict
    self._sketch = sketch

  def __setitem__(self, k, v):
    self._sketch.setDatumWrapped(self._indexDict[k], v)


class FreecadProperty:
  '''
  FreecadProperty represents a property of an object in the document tree and can 
  be created from normal python shells without any connection to FreeCAD. It can 
  be read from and written to as if it was the real freecad property. All 
  read/writes will be forwarded to the running Freecad child process.
  '''
  def __init__(self, doc, obj, path, isCall=False, internalObjectName=False):
    self.__dict__['_doc'] = doc
    self.__dict__['_obj'] = obj
    self.__dict__['_path'] = path
    self.__dict__['_internalObjectName'] = internalObjectName
    self.__dict__['_isConstraint'] = False
    self.__dict__['_sketchObjReference'] = None
    self._ensureExists(isCall=isCall)

  def __repr__(self):
    return self.__str__()

  def __str__(self):
    return f'<FreecadProperty {self._obj}.{self._path}, value: {self.getStr()}>'

  def _freecadShellRepr(self):
    if self._internalObjectName:
      return f'App.activeDocument().getObject("{self._obj}").{self._path}'    
    return f'App.activeDocument().getObjectsByLabel("{self._obj}")[0].{self._path}'

  def _ensureExists(self, isCall=False):
    # log in case of function call
    if isCall:
      if _PRINT_SETTER_AND_CALL_LINES:
        io.verb(f'running call line: {self._freecadShellRepr()}')
      pass
    
    # if fast mode is enabled, skip any existCheck and run calls without waiting for response
    if self._doc._fastModeEnabled():
      if isCall:
        self._doc.write(self._freecadShellRepr())

    # run proper query and error checking if not in fast mode
    else:
      try:
        rn = f'{random.random():.8f}'
        self._doc.query(f'{self._freecadShellRepr()}; print("success{rn}")', expect=f'success{rn}',
                        errText=f'property {self} does not exist')
      except Exception:
        io.warn(f'running {"call" if isCall else "ensureExists"} line failed: {self._freecadShellRepr()}')
        raise

  # ----------------------------------
  # ITEM AND ATTRIBUTE ACCESS

  def set(self, value, lvalSuffix='', **kwargs):
    if self._isConstraint:
      return self.setDatumWrapped(self._constraintIndex, value)
    else:
      setterLine = f'{self._freecadShellRepr()}{lvalSuffix} = {repr(value)}'
      if _PRINT_SETTER_AND_CALL_LINES:
        io.verb(f'running setter line: {setterLine.strip()}')

      # if fast mode is enabled run setter line without waiting for response
      if self._doc._fastModeEnabled():
        self._doc.write(setterLine)

      # run proper query and error checking if not in fast mode
      else:
        rn = f'{random.random():.8f}'
        self._doc.query(f'{setterLine}; print("success{rn}")', expect=f'success{rn}',
                        errText=f'failed running python line in FreeCAD: {setterLine.strip()}')

  def setDatumWrapped(self, constraintIndex, v):
      # select good unit for value
      if abs(v) < 1e-6:
        unit = 'nm'
        _v = 1e6*v
      elif abs(v) < 1e-3:
        unit = 'um'
        _v = 1e3*v
      else:
        unit = 'mm'
        _v = v
      # return expression
      valueExpr = FreecadExpression(f'App.Units.Quantity("{_v:.6f} {unit}")')
      
      # force careful flush before and after setDatum, because strange "invalid constraint index" may occur otherwise
      for retry in range(99):
        result = self._sketchObjReference.setDatum(constraintIndex, valueExpr)
        self._doc._flushOutput(forceCareful=True, keepErrs=True)
        if err:=self._doc.readErr():
          if retry > 5:
            raise RuntimeError(f'setting datum {constraintIndex} of {self} failed: {err}')
          io.warn(f'error {err} occurred while setting a datum, retrying...')
          self._doc._flushOutput(forceCareful=True)
          time.sleep(0.1*2**retry)
        else:
          break
      return result

  def __setattr__(self, key, value):
    self.set(value, lvalSuffix=f'.{key}')

  def __getattr__(self, key):
    return FreecadProperty(self._doc, self._obj, f'{self._path}.{key}', 
                           internalObjectName=self._internalObjectName)

  def __setitem__(self, key, value):
    self.set(value, lvalSuffix=f'[{repr(key)}]')

  def __getitem__(self, key):
    return FreecadProperty(self._doc, self._obj, f'{self._path}[{repr(key)}]', 
                           internalObjectName=self._internalObjectName)

  def __len__(self):
    return int( self._doc.query( f'len( {self._freecadShellRepr()} )' ) )

  def __iter__(self):
    return iter([self[i] for i in range(len(self))])

  def __call__(self, *args, **kwargs):
    argsStr = ''
    if len(args):
      argsStr += f'*{args}'
    if len(kwargs):
      if len(argsStr):
        argsStr += ', '
      argsStr += f'**{kwargs}'
    p = FreecadProperty(self._doc, self._obj, f'{self._path}({argsStr})', isCall=True, 
                        internalObjectName=self._internalObjectName)

  def getStr(self):
    if self._isConstraint:
      return self.Value.getStr()
    return self._doc.query(self._freecadShellRepr())

  def getFloat(self):
    return float(self.getStr())
  
  def getInt(self):
    return float(self.getStr())

  def get(self):
    _str = self.getStr()
    try:
      return eval(_str)
    except Exception:
      return _str

  # ----------------------------------
  # CUSTOM ATTRIBUTES AND METHODS
  def markAsConstraint(self, sketchObj, sketchIndex):
    self.__dict__['_isConstraint'] = True
    self.__dict__['_sketchObjReference'] = sketchObj
    self.__dict__['_constraintIndex'] = sketchIndex
    return self

  def getConstraintsByName(self):
    indexAndConstraints = { c.Name.get().strip(): (i, c.markAsConstraint(self, i))
                                            for i, c in enumerate(self.Constraints)
                                                              if c.Name.get().strip() }
    return FreecadConstraintDict(constraintDict={k:v[1] for k,v in indexAndConstraints.items()},
                                 indexDict={k:v[0] for k,v in indexAndConstraints.items()},
                                 sketch=self)


class FreecadObject(FreecadProperty):
  '''
  FreecadObject represents an object from the document tree and can be created from
  normal python shells without any connection to FreeCAD. It can be read from and
  written to as if it was the real freecad object. All read/writes will be forwarded
  to the running Freecad child process.
  '''
  def __init__(self, doc, obj, internalObjectName=False):
    self.__dict__['_doc'] = doc
    self.__dict__['_obj'] = obj
    self.__dict__['_internalObjectName'] = internalObjectName
    self._ensureExists()

  def __str__(self):
    return f'<FreecadObject {os.path.basename(self._doc._path)}: {self._obj}>'

  def _freecadShellRepr(self):
    if self._internalObjectName:
      return f'App.activeDocument().getObject("{self._obj}")'
    return f'App.activeDocument().getObjectsByLabel("{self._obj}")[0]'

  def _ensureExists(self):
    # ensure at exactly one object exists
    if not self._internalObjectName:
      try:
        # force careful flushing because we need the exact response to object count also in fastMode
        self._doc._flushOutput(forceCareful=True)
        objectCount = int(self._doc.query(f'print(len( App.activeDocument().getObjectsByLabel("{self._obj}") ))'))
        # zero objects with given label exist -> fail
        if objectCount == 0:
          raise ValueError(f'object with label {self._obj} does not exist.')
        # more than one object with given label exist -> fail
        if objectCount != 1:
          raise ValueError(f'object label {self._obj} is not unique, consider to rename it '
                           f'or to use the internal name instead.')

      # handle value errors by retrying with "internal name" style instead of label
      except ValueError:
        self.__dict__['_internalObjectName'] = True
        self._ensureExists()

    # make sure self._freecadShellRepr() evaluates fine
    try:
      queryResponse = self._doc.query(f'print({self._freecadShellRepr()})')
      if queryResponse == 'None':
        raise ValueError(f'expected freecad object, found None')
    except Exception:
      raise ValueError(f'object with {"internal name" if self._internalObjectName else "label"} {self._obj} does '
                       f'not exist in freecad document (see exceptions above for detailed reason)')
  
  def __setattr__(self, key, value):
    self.__getattr__(key).set(value)

  def __getattr__(self, key):
    return FreecadProperty(self._doc, self._obj, key, 
                           internalObjectName=self._internalObjectName)


class FreecadDocument:
  '''
  FreecadDocument class represents a FCStd document and can be created from normal 
  python shells without any connection to FreeCAD. The FCStd file properties can
  be manipulated and optical simulations can be started as if we were running within
  the FreeCAD-shell.
  
  The document is intended to be used as a context manager to clean up after itself.
  '''
  def __init__(self, path=None, openFreshCopy=False):
    # register self in global list
    _ALL_DOCUMENTS.append(self)

    # autodetect path is none is given
    if path is None:
      _path = os.path.realpath(os.getcwd())
      while _path.count('/') != 1:
        if _path.endswith('.OpticsDesign'):
          candidate = _path[:-13]+'.FCStd'
          if os.path.exists(candidate):
            path = candidate
            break
        _path = os.path.dirname(_path)
      else:
        raise RuntimeError('failed to autodetect path of FCStd project file')
    if not os.path.exists(path):
      raise ValueError(f'path to freecad project file {path} does not seem to exist')
    if not path.endswith('.FCStd'):
      raise ValueError(f'path to freecad project file {path} does end with .FCStd')

    # if openFreshCopy is True, create temp folder, copy freecad file in there, 
    # and modify path attributes accordingly
    self._openFreshCopy = openFreshCopy
    self._originalPath = path
    if openFreshCopy:
      # find unique filename for FCStd file
      tempDirName = self._getTempFolder()
      while True:
        uniqueName = f'{os.path.basename(path)[:-6]}-{int(time.time()*1e3)}-pid{os.getpid()}-thread{threading.get_ident()}.FCStd'
        if not os.path.exists(f'{tempDirName}/{uniqueName}'):
          break

      # copy FCStd file to temp dir
      _target = f'{tempDirName}/{uniqueName}'
      os.makedirs(os.path.dirname(_target), exist_ok=True)
      shutil.copy(path, _target)

      # modify path variable for following initialization
      path = f'{tempDirName}/{uniqueName}'

    # save in attributes
    self._path = path

    # generate results folder path
    self._resultsPath = path[:-6]+'.OpticsDesign'

    # setup flags
    self._isquit = False
    self._isterminate = False
    self._iskill = False
    self._isRunning = False
    self._freecadInteractionTimesList = [time.time()]
    self._previousFastModeEnabled = False

  def __repr__(self):
    return self.__str__()

  def __str__(self):
    return f'<FreecadDocument {os.path.basename(self._path)}>'

  def resultsPath(self):
    return self._resultsPath

  def _getTempFolder(self, create=False):
    tempDirName = f'{self._originalPath[:-6]}.OpticsDesign/tmp'
    os.makedirs(tempDirName, exist_ok=True)
    return tempDirName

  def purgeTempFolder(self):
    # dont clean if this is a tmp-document
    if self._openFreshCopy:
      raise ValueError(f'this freecad document was opened using openFreshCopy=True, '
                       f'can only purge temp folder from instances that were opened '
                       f'without openFreshCopy option')
    shutil.rmtree(self._getTempFolder())

  def isOpenFreshCopy(self):
    if self._openFreshCopy:
      return True
    return '.OpticsDesign/tmp'.lower() in self._path.lower()

  def _sanitizeTempFolder(self, timeout=4):
    # dont clean if this is a tmp-document
    if self.isOpenFreshCopy():
      return

    t0 = time.time()
    _tmp = self._getTempFolder(create=False)
    lastWarn = time.time()
    if os.path.exists(_tmp):

      # retry cleanup on fail
      cleanedSomething = True 
      while cleanedSomething:
        maxLevel = 3 # <- recurse down to tmp/...OpticsDesign/raw/sim... level but not deeper
        try:
          cleanedSomething = False

          # walk through tmp-dir tree and delete anything that 
          # did not change for more than a day, remove empty
          # directories
          for r, dirs, files in os.walk(_tmp, topdown=True):
            # remove empty dirs
            for d in dirs:
              if not len(os.listdir(f'{r}/{d}')):
                os.rmdir(f'{r}/{d}')
                cleanedSomething = True

            # remove dirs on third sublevel by age
            dirLevel = r.count('/') - _tmp.count('/') 
            if dirLevel >= maxLevel:
              for d in dirs:
                maxAge = 24*60*60
                if time.time()-os.stat(f'{r}/{d}').st_mtime > maxAge:
                  shutil.rmtree(f'{r}/{d}')
                  cleanedSomething = True
            
            # do not recurse deeper then to third sublevel
            if dirLevel >= maxLevel:
              dirs[:] = []
            
            # remove old files
            for f in files:
              #io.verb(f'checking {dirLevel=} {r}/{f}')
              maxAge = 24*60*60
              if f.endswith('FCStd'):
                maxAge = 7*24*60*60
              try:
                if time.time()-os.stat(f'{r}/{f}').st_mtime > maxAge:
                  os.remove(f'{r}/{f}')
                  cleanedSomething = True
              except Exception as e:
                pass

            # apologize if cleaning takes long
            if time.time()-lastWarn > 5:
              lastWarn = time.time()
              io.warn(f'sorry, house keeping of the tmp-folder '
                      f'is taking a lot of time... (cleaning since {io.secondsToStr(time.time()-t0)})')

            # abort cleaning if time is up
            if time.time()-t0 > timeout:
              io.verb(f'tmp folder clean timed out after {io.secondsToStr(time.time()-t0)}')
              return
        except Exception:
          io.warn(f'exception raised during tmp-dir cleanup:\n\n'+traceback.format_exc())
      io.verb(f'full tmp folder clean successful after {io.secondsToStr(time.time()-t0)}')

  ###########################################################################
  # FILE MANIPULATION/SIMULATION TRIGGER LOGIC

  def __getattr__(self, name):
    return self.getObject(name)

  def getObject(self, name):
    return FreecadObject(self, name)

  def objects(self, internalNames=False):
    if internalNames:
      return sorted(list(set(eval(self.query(f'[o.Name for o in App.activeDocument().Objects]')))))
    return sorted(list(set(eval(self.query(f'[o.Label for o in App.activeDocument().Objects]')))))

  def runSimulation(self, action='true', endIf=None, endIfMaxLoad=.5):
    '''
    Start a simulation. 
    
    The action argument specifies the simulation kind and has to be one of
    "true", "singletrue", "pseudo", "singlepseudo" or "fans". 
    
    The endIf argument allows to pass a custom callback which will be executed
    continuously during the simulation with the current result folder path as
    its argument.

    The endIfMaxLoad argument must be between 0.01 and 1 and specifies which 
    percentage of the overall time the endIf callback is allowed to run on 
    average. The mainloop will add delays between endIf(...)-calls. In any
    case, the endIf(..) will run not more often than once per second, and at
    least once per hour.
    '''
    try:
      # behave as slave process while simulation is running (the FreeCAD simulation master
      # running in the background wants to be master, therefore we have to step back)
      processes.jupyterBecomeSlave()

      # check actions argument
      allowedActions = 'true pseudo singletrue singlepseudo fans'.split()
      if action not in allowedActions:
        raise ValueError(f'illegal action {action}, expected one of {", ".join(allowedActions)}')

      # save list of existing raw data folders
      rawFoldersBefore = []
      if os.path.exists(self._resultsPath+'/raw'):
        rawFoldersBefore = os.listdir(self._resultsPath+'/raw')

      # subfunction to detect new results folder
      def newFolder(allowEmpty=True):
        # find and return newly appeared folder in results dir
        newFolders = []
        if os.path.exists(self._resultsPath+'/raw'):
          newFolders = [d for d in os.listdir(self._resultsPath+'/raw')
                                                if d not in rawFoldersBefore]

        # check plausibility of result
        if len(newFolders) == 0 and not allowEmpty:

          # no new folder appeared, show different exception texts depending on 
          # fans/single or if error output occurred
          if action == 'fans' or 'single' in action:
            raise RuntimeError(f'no new results folder was created during {action} '
                              f'simulation, did you enable "Enable Store Single Shot '
                              f'Data" in the active Simulation Settings?'
                              +self._errorOutText())
          raise RuntimeError(f'no new results folder was created during {action} '
                             f'simulation, is another simulation running? '
                             f'{self._path=}, {self._resultsPath=}'
                             +self._errorOutText())
        
        # more than one new folder appeared
        elif len(newFolders) > 1:
          raise RuntimeError(f'somehow more than one result folder was created '
                            f'during {action} simulation, this should not be '
                            f'possible...'
                            +_errorOutText())
        # return result
        return RawFolder(f'{self._resultsPath}/raw/{newFolders[0]}') if len(newFolders) else None 

      # write command
      self.write(f'from freecad.optics_design_workbench import simulation',
                 f'simulation.runSimulation(action="{action}")')

      # start progress tracker background thread
      progressTracker = progress.progressTrackerInstance(doc=self)

      # start mainloop that detects end of simulation and endIf-calls
      possibleStacktrace = []
      try:
        lastSentRandom = 0
        lastEndIfCheck = time.time()
        endIfDuration = 0
        rn, l = None, None
        while rn is None or rn != l:
          # ask to print random number every 60 seconds 
          if time.time()-lastSentRandom > 60:
            rn = f'{random.random():.8f}'
            self.write(f'print("{rn}")')
            lastSentRandom = time.time()

          # check for new folders and update simulation tracker folder if so
          _newFolder = newFolder()
          progressTracker.resultsFolder = _newFolder

          # check custom simulation-end criterion with ~50% dutycycle
          if ( endIf is not None 
                and time.time()-lastEndIfCheck
                      > min([60*60, max([1, 1/max([0.01, endIfMaxLoad]) * endIfDuration ]) ]) ):
            lastEndIfCheck = time.time()
            if _newFolder is not None:
              # if endIf callback returns True, place simulation-done file
              _t0 = time.time()
              if endIf(_newFolder):
                with open(f'{self._resultsPath}/simulation-is-done', 'w') as _:
                  pass
              endIfDuration = time.time()-_t0

          # if random number appears in output the simulation must be done
          for l in self.read():
            if rn and rn == l:
              break

            # if line looks like the beginning of a stacktrace, reset stacktrace buffer
            if 'traceback (most recent call last)' in l.lower():
              possibleStacktrace = []
            possibleStacktrace.append(l)

          # limit loop speed
          time.sleep(1e-2)

        # presence of simulation-canceled file or absence of simulation-is-done file indicates
        # that simulation was ended by an exception: raise this exception
        if (os.path.exists(f'{self._resultsPath}/simulation-is-canceled') 
                or not os.path.exists(f'{self._resultsPath}/simulation-is-done')):
          stacktraceGuess = ''
          if len((''.join(possibleStacktrace)).strip()):
            stacktraceGuess = f'; I guess it has something to do with this output:\n\n'+'\n'.join(possibleStacktrace)
          logfile = os.path.relpath(self._resultsPath+'/optics_design_workbench.log')
          raise RuntimeError(f'simulation process failed, see logfile {logfile} for details'
                              +stacktraceGuess
                              +self._errorOutText())

        # return new results folder generated during the simulation
        return newFolder(allowEmpty=False)

      # clean up progress tracker
      finally:
        # increment progress tracker iteration count
        progressTracker.nextIteration()

    # let jupyter become master again after simulation is done
    finally:
      processes.jupyterBecomeMaster()

  ###########################################################################
  # LAUNCH LOGIC

  def __enter__(self):
    self.open()
    return self

  def open(self):
    self._updateInteractionTime()
    t0 = time.time()
    
    # register that this process is a jupyter process to setup logging
    processes.setupJupyterMaster(self._path)
    io.verb(f'opening {self} instance...')

    # signalizing progress tracker module that it is now allows to create
    # a global progress tracker
    progress.ALLOW_PROGRESS_TACKERS = True

    # launch child process - DONT load file here or it will be loaded without 
    #                        ViewProvider objects because GUI is not yet up
    self._p = subprocess.Popen([_DEFAULT_FREECAD_EXECUTABLE, '-c'],
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE,
                                stdin=subprocess.PIPE, 
                                text=True, bufsize=-1)
    self._isRunning = True

    # start thread the continuously reads stdout and adds results to queue
    self._q = queue.Queue()
    self._qe = queue.Queue()
    def readOutput():
      for line in iter(self._p.stdout.readline, ''):
        if line.strip():
          # remove line ending characters
          while line.endswith('\r') or line.endswith('\n'):
            line = line[:-1]
          #print(f'received line {line}')
          self._q.put(line)
        #time.sleep(1e-6)
      self._p.stdout.close()
    self._t = threading.Thread(target=readOutput)
    self._t.daemon = True # thread dies with the program
    self._t.start()

    def readError():
      for line in iter(self._p.stderr.readline, ''):
        # remove prompt characters from beginning of line
        while line.lstrip().startswith('>>>') or line.lstrip().startswith('...'):
          line = line.lstrip()[4:]

        # not-important-list: ignore some well known errors which are rather warnings than errors:
        if any([p.lower() in line.lower() for p in (
              'Updating geometry: Error build geometry',
              'Invalid solution from', )]):
          io.warn(f'ignoring FreeCAD error output {repr(line.strip())}')
          line = ''

        # very-important-list: raise certain errors immediately, as they indicate a broken document state:
        if any([p.lower() in line.lower() for p in (
             'BRep_API: command not done', 'Revolution: Revolve axis intersects the sketch',)]):
          raise RuntimeError(f'FreeCAD reported error: {line}')

        if line.strip():
          # remove line ending characters and add to queue
          while line.endswith('\r') or line.endswith('\n'):
            line = line[:-1]
          io.warn(f'received error line {repr(line)}', logOnly=True)
          self._qe.put(line)
        time.sleep(1e-3)
      self._p.stdout.close()
    self._te = threading.Thread(target=readError)
    self._te.daemon = True # thread dies with the program
    self._te.start()

    # Send python snippet to load GUI and immediately hide window, this allows to run
    # in a "headless-like" mode but still loads all ViewProvider objects. In true headless
    # mode without the ViewProvider objects, saving the file will break many things, because
    # all View Properties are lost then. Exit on error to make failure visible to this
    # process (we dont read stdout/stderr and only realize something is wrong if the child dies)
    # The second half of the snippets makes sure all other gui windows the workbench might want
    # to open remain hidden (or do not open at all)
    self.write(f'try: ',
               f'  import FreeCADGui as Gui',
               f'  Gui.showMainWindow()',
               f'  Gui.getMainWindow().hide()',
               f'  ',
               f'  from freecad.optics_design_workbench.gui_windows import common',
               f'  common.HIDE_GUI = True',
               f'except Exception:',
               f'  exit()')

    # flush outputs with infinite timeout to make sure startup was completely done
    self._flushOutput(forceCareful=True)

    # after gui is fully loaded, load document
    self.write(f'try: ',
               f'  App.openDocument({repr(self._path)})',
               f'except Exception:',
               f'  exit()')

    # flush outputs with infinite timeout to make sure loading file was completely done
    self._flushOutput(forceCareful=True)

    # make sure numpy is imported globally, as numpy and as np, to ensure all numpy datatypes
    # can be constructed properly
    self.write(f'import numpy',
               f'import numpy as np',
               f'from numpy import *')

    # flush outputs with infinite timeout to ensure this was successfully imported
    self._flushOutput(forceCareful=True)

    # clear interaction time list to avoid immediately triggering fast mode
    self._forceDisableFastMode()

    # print success
    io.verb(f'done in {time.time()-t0:.1f}s')


  ###########################################################################
  # SUBPROCESS IO LOGIC

  def _errorOutText(self):
    'return a string to append to exceptions which hints to freecad error output'
    errorOut = self.readErr()
    if not errorOut.strip():
      return ''
    return (f'\n\nFreecad process had possibly related '
            f'error output recently:\n\n{errorOut}')

  def lastInteractionTime(self):
    return self._freecadInteractionTimesList[-1]

  def _updateInteractionTime(self):
    self._freecadInteractionTimesList.append(time.time())
    T = array(self._freecadInteractionTimesList)
    self._freecadInteractionTimesList[:] = T[time.time()-T < 30]

  def _forceDisableFastMode(self):
    self._freecadInteractionTimesList[:] = [time.time()]

  def _fastModeEnabled(self):
    T = array(self._freecadInteractionTimesList)

    # enable fast mode if more than 100 freecad interactions
    # within last 3 seconds, only disable if less then 100
    # interactions within 10 seconds 
    #io.verb(f'{sum( time.time()-T < 5 )=}')
    if self._previousFastModeEnabled:
      enable = sum( time.time()-T < 10 ) > 100
    else:
      enable = sum( time.time()-T < 3 ) > 100

    # if fast mode was just ended, make sure to flush buffers
    if not enable and self._previousFastModeEnabled:
      self._flushOutput(forceCareful=True)

    # update previousFastMode attribute and log if fastModeEnable changed
    if enable != self._previousFastModeEnabled:
      io.verb(f'{"enabled" if enable else "disabled"} freecad communication fast mode')
      self._previousFastModeEnabled = enable

    return enable

  def write(self, *data):
    self._updateInteractionTime()
    cmdStr = '\r\n'+'\r\n'.join(data)+'\r\n'*2 # add plenty of newlines at the and to
                                               # make sure command is complete also 
                                               # if indented a few levels
    if _PRINT_FREECAD_COMMUNICATION:
      io.verb('> '+cmdStr.replace('\r', '').strip('\n'))
    self._p.stdin.write(cmdStr)
    self._p.stdin.flush()

  def _flushOutput(self, timeout=60, forceCareful=False, keepErrs=False):
    self._updateInteractionTime()
    t0 = time.time()

    # throw away any previous content
    if not keepErrs:
      self.readErr()
    self.read()

    # skip careful flush if fastMode is active (and not forced)
    if forceCareful or not self._fastModeEnabled():
      lastWarned = time.time()
      lastPrintedRn = 0
      rn = None
      while True:
        # ask to print random number every few seconds 
        # (scale wait time with time that has passed)
        if time.time()-lastPrintedRn > 1/5*(time.time()-t0):
          lastPrintedRn = time.time()
          rn = f'{random.random():.8f}'
          self.write(f'print("{rn}")')
        
        # if random number appears in output: success
        out = self.read()
        #print(f'waiting for {rn}, found {out}')
        if rn in out:
          return

        # warn of takes long
        if time.time()-lastWarned > 5:
          lastWarned = time.time()
          io.warn(f'long waiting time for freecad process to become responsive, '
                  f'waiting since {io.secondsToStr(time.time()-t0)} '
                  f'(this may happen if the system is under heavy load)')

        # if time is up: raise timeout error
        if time.time()-t0 > timeout:
          raise RuntimeError(f'failed to flush output buffer of FreeCAD, '
                            f'is a process featuring heavy output printing '
                            f'running?')
        time.sleep(1e-3)

  def read(self):
    self._updateInteractionTime()
    result = []
    try:
      while True:
        result.append(self._q.get_nowait())
    except queue.Empty:
      pass
    return result

  def readErr(self):
    self._updateInteractionTime()
    result = []
    lastLineReceived = 0
    while True:
      # fetch lines until queue.Empty is raised
      try:
        while True:
          result.append(self._qe.get_nowait())
          lastLineReceived = time.time()
      except queue.Empty:
        pass

      # if a line was received once only exit after
      # no line was seen for 500ms
      if time.time()-lastLineReceived > .5:
        break

      # limit loop seed
      time.sleep(1e-3)

    return '\n'.join(result)

  def readline(self):
    self._updateInteractionTime()
    try:
      return self._q.get_nowait()
    except queue.Empty:
      pass
    return None

  def query(self, *data, timeout=60, expect=None, errText=None):
    self._updateInteractionTime()
    t0 = time.time()

    # make sure output buffer of freecad is empty (do not wait
    # full timeout to leave a bit time for the query itself)
    self._flushOutput(timeout=0.9*timeout)

    # send python commands
    self.write(*data)

    # wait for single line response
    lastWarned = time.time()
    sentCommandT0 = time.time()
    while True:
      if error:=self.readErr():
        raise RuntimeError((f'{errText.strip()}\n' if errText else '')
                           +f'exception was raised while handling '
                           +f'command(s) {data}:\n\n'+error)

      # check for result line
      if line:=self.readline():
        if _PRINT_FREECAD_COMMUNICATION:
          io.verb(f'< {line}')
        if expect is not None and expect not in line:
          continue
        return line

      # warn of takes long
      if time.time()-lastWarned > 5:
        lastWarned = time.time()
        io.warn(f'long waiting time for response from freecad process, waiting '
                f'since {io.secondsToStr(time.time()-sentCommandT0)} '
                f'(this may happen if the system is under heavy load)')

      # raise if timeout has passed
      if time.time()-t0 > timeout:
        raise RuntimeError((f'{errText.strip()}\n' if errText else '')
                           +f'failed to fetch any response to command(s) {data}, '
                           +f'is the system overloaded or is freecad running '
                           +f'something with heavy output printing running?')
      time.sleep(1e-3)

  ###########################################################################
  # SHUTDOWN LOGIC

  def __exit__(self, *args, **kwargs):
    self.close()

  def save(self):
    # send clear rays command
    self.write('Gui.runCommand("Clear all rays",0)')

    # send save document command
    self.write('App.activeDocument().save()')

  def close(self):
    t0 = time.time()
    io.verb(f'closing {self} instance...')

    # save changes to disk (if self is open)
    if self.isRunning():
      self.save()

    while self.isRunning():
      # gently ask to quit
      self._quit()

      # if not exited after three seconds, send sigterm
      if time.time()-t0 > 3:
        self._terminate()

      # if not exited after five seconds, send sigkill
      if time.time()-t0 > 5:
        self._kill()

      # limit loop speed
      time.sleep(.1)

    # make sure global progress tracker instance is shut down
    # and prevent creating new ones because we are not in the 
    # the document context manager anymore
    if progress.progressTrackerExists():
      progress.progressTrackerInstance(doc=self).quit()
      progress.ALLOW_PROGRESS_TACKERS = False

    # run temp dir cleanup
    self._sanitizeTempFolder()

    io.verb(f'done in {time.time()-t0:.1f}s')
    processes.unsetJupyterMaster()

  def isRunning(self):
    if self._isRunning:
      if (res:=self._p.poll()) is not None:
        self._isRunning = False
        io.verb(f'FreeCAD finished (exit code {res})')
    return self._isRunning

  def _quit(self):
    if self.isRunning() and not self._isquit:
      io.verb('asking FreeCAD to quit...')
      self.write('FreeCADGui.getMainWindow().deleteLater()') # <- trick to gently shut down freecad without any prompts
      #self.write('exit()') # <- dont do, exiting the python shell causes segfaults sometimes
      self._isquit = True

  def _terminate(self):
    if self.isRunning():
      if not self._isterminate:
        io.verb('terminating FreeCAD...')
        self._isterminate = True

      # place cancel file
      with open(f'{self._resultsPath}/simulation-is-canceled', 'w') as f:
        pass
      try:
        os.remove(f'{self._resultsPath}/simulation-is-running')
      except Exception:
        pass

      # close stdin and send sigterm
      try:
        self._p.stdin.close()
      except:
        pass
      self._p.send_signal(signal.SIGTERM)

  def _kill(self):
    if self.isRunning():
      if not self._iskill:
        io.verb('killing FreeCAD...')
        self._iskill = True
      self._p.send_signal(signal.SIGKILL)


@functools.wraps(FreecadDocument.__init__)
def openFreecadGui(*args, **kwargs):
  # close all sweepers that may have opened the file
  parameter_sweeper.closeAllSweepers()

  # create dummy-document object to detect all necessary paths 
  document = FreecadDocument(*args, **kwargs)

  # launch freecad process
  p = subprocess.Popen([_DEFAULT_FREECAD_EXECUTABLE, document._path],
                        stdout=subprocess.DEVNULL, 
                        stderr=subprocess.DEVNULL)
  _didFinishPrint = False
  def isRunning():
    nonlocal _didFinishPrint
    if (res:=p.poll()) is not None:
      if not _didFinishPrint:
        io.verb(f'FreeCAD finished (exit code {res})')
        _didFinishPrint = True
      return False
    return True

  # block until freecad is closed, kill freecad process if ctrl+c is caught
  while isRunning():
    try: 
      # limit loop speed
      time.sleep(1/3)

    # if ctrl+c is received make sure to kill freecad
    except KeyboardInterrupt:
      # first gently...
      for _ in range(5):
        p.send_signal(signal.SIGTERM)
        time.sleep(1/2)
        if not isRunning():
          break

      # ..then not so gently
      while isRunning():
        p.send_signal(signal.SIGKILL)
        time.sleep(1/2)

      # re-raise Keyboard interrupt because absorbing a keyboard interrupt
      # is not a good idea
      raise

def _rawFolders(basePath='.'):
  # descent path until 'raw' folder is found
  while not os.path.exists(f'{basePath}/raw') and basePath != '/':
    basePath = os.path.realpath(f'{basePath}/..')
  basePath = f'{basePath}/raw'
  if not os.path.exists(basePath):
    raise ValueError(f'failed to find "raw" folder in any parent directory of {basePath=}')

  # get all simulation run folders
  folders = sorted([d for d in os.listdir(basePath) if d.startswith('simulation-run-')])
  indices = [int(d[len('simulation-run-'):]) for d in folders]
  return basePath, folders, indices

def rawFolders(basePath='.'):
  basePath, folders, indices = _rawFolders(basePath=basePath)
  return RawFolderRange( [os.path.relpath(f'{basePath}/{f}') for f in folders] )

def rawFolderByIndex(index=-1, basePath='.'):
  basePath, folders, indices = _rawFolders(basePath=basePath)

  # interpret positive indices just like number in the directory name
  if index >= 0:
    if index not in indices:
      raise ValueError(f'simulation-run folder with index {index} does not exist')
    return RawFolder(f'{basePath}/{folders[indices.index(index)]}')

  # interpret negative indices as counting from the end (ignoring directory name)
  return RawFolder(f'{basePath}/{folders[index]}')

@functools.wraps(rawFolderByIndex)
def latestRawFolder(**kwargs):
  return rawFolderByIndex(index=-1, **kwargs)

def _updateResultEntry(result, key, value):
  v = value
  k = key

  # key not yet existing: just save
  if k not in result.keys():
    result[k] = v

  # string + string = string if identical, else list of two strings
  elif type(v) in (str, str_) and type(result) in (str, str_):
    if v == result[k]:
      pass
    else:
      result[k] = [result[k], v]

  # string + list = add string to list if not existing yet
  elif type(v) in (str, str_) and type(result) not in (str, str_) and hasattr(result[k], '__iter__'):
    if v not in result[k]:
      result[k] = list(result[k]) + [v]

  # list/array + list/array = concatenate
  elif len(result[k]) == 0:
    result[k] = v
  elif len(v) == 0:
    pass
  else:
    result[k] = concatenate([result[k], v], axis=0)


class RawFolder:
  '''
  RawFolder class represents one simluation-run folder in the raw subfolder of
  the simulation results directory.
  '''
  def __init__(self, path, timeout=60):
    self._path = path

    # detect uid, wait up to timeout for uid file to show up
    t0 = time.time()
    lastWarned = time.time()
    while True:
      candidates = [f for f in os.listdir(self._path) 
                        if os.path.isfile(f'{self._path}/{f}') and f.startswith('uid')]
      # if exactly one candidate: save uid and exit loop
      if len(candidates) == 1:
        self._uid = candidates[0][4:] 
        break

      # warn if it takes long
      if time.time()-lastWarned > 3:
        lastWarned = time.time()
        io.warn(f'waiting for uid file to show up took long ({io.secondsToStr(time.time()-t0)})')

      # raise after timeout
      if time.time()-t0 > timeout:
        if len(candidates) == 0:
          raise RuntimeError('invalid raw data folder: uid file missing')
        if len(candidates) > 1:
          raise RuntimeError('invalid raw data folder: more than one uid file')

  def __repr__(self):
    return self.__str__()

  def __str__(self):
    return f'<RawFolder {os.path.basename(self._path)}/ UID={self._uid}>'

  def path(self):
    return os.path.relpath(self._path)

  def reload():
    self.tree.cache_clear()

  @functools.cache
  def tree(self, _path=None):
    '''
    Return nested directories representing the directory structure of this RawFolder
    '''
    if _path is None:
      _path = self._path
    result = {}
    folderContents = []
    for d in os.scandir(_path):
      # add subdirectory as subdictionary
      if d.is_dir():
        result[d.name] = self.tree(_path=d.path)

      # register presence of known and unknown files
      elif d.name.endswith('-hits.pkl'):
        folderContents.append('hit')
      elif d.name.endswith('-rays.pkl'):
        folderContents.append('ray')
      elif d.name.startswith('uid-') or d.name.startswith('.'):
        pass
      else:
        folderContents.append('unknown')

    # add simple string entries in dictionary for detected files
    for found in set(folderContents):
      result[f'<{folderContents.count(found)} {found} files>'] = None
    return result

  def printTree(self, _node=None, _prefix='  '):
    '''
    Pretty print the directory structure of this RawFolder
    '''
    if _node is None:
      print(f'{os.path.basename(self._path)}/')
      _node = self.tree()
    for k, v in sorted(_node.items(), key=lambda e: e[0]):
      if v is None:
        print(_prefix+k)
      elif type(v) is dict:
        print(_prefix+k+'/')
        self.printTree(_node=v, _prefix=_prefix+'  ')
      else:
        raise ValueError(f'invalid tree node {_node}')

  def loadHits(self, pattern='*'):
    return hits.Hits(self._load(pattern=pattern, kind='hits'))

  def loadRays(self, pattern='*'):
    return self._load(pattern=pattern, kind='rays')

  def _findPathsAndSanitize(self, pattern, kind, optimalFilesize=500e6):
    if pattern == '*':
      pattern = '**'
    def _makeGlob():
      return glob.iglob(f'{self._path}/{pattern}/*-{kind}.pkl', recursive=True)

    # group all files to consider by folder
    byFolder = {}
    for path in _makeGlob():
      base, name = os.path.split(path)
      if base not in byFolder.keys():
        byFolder[base] = []
      try:
        _stat = os.stat(path)
        mtime = _stat.st_mtime
        size = _stat.st_size
      except Exception:
        mtime = time.time()
        size = 1
      byFolder[base].append([name, mtime, size])

    # find files that look well suited for merging
    for base, namesTimesSizes in byFolder.items():
      mergeList = []
      sizeInList = 0
      def mergeAllInList():
        # only do something if more than one file in list
        if len(mergeList) > 1:
          # load and merge all files in list
          merged = {}
          for name in mergeList:
            try:
              with open(f'{base}/{name}', 'rb') as _f:
                data = pickle.load(_f)
            except Exception as e:
              io.warn(f'failed to read {kind} file {base}/{name}: {e.__class__.__name__} "{e}"')
            else:
              # merge file content with merged dict
              for k, v in data.items():
                _updateResultEntry(merged, k, v)
          
          # overwrite first file in list with total results
          with atomic_write(f'{base}/{mergeList[0]}',
                            mode='wb', overwrite=True) as f:
            pickle.dump(merged, f)

          # delete all remaining files in list
          for name in mergeList[1:]:
            os.remove(f'{base}/{name}')

        # reset list
        mergeList.clear()

      # only consider files that did not change in the last hour
      for name, size in sorted([(n, s) for n, t, s in namesTimesSizes if time.time()-t > 60*60],
                               key=lambda e: e[0]):
        # if this file would make the current merge collection much larger than optimal
        # size -> merge without current file
        if sizeInList+size > 1.5*optimalFilesize:
          mergeAllInList()

        # add current file to list and merge if collection is larger than optimal size
        mergeList.append(name)
        sizeInList += size
        if sizeInList > optimalFilesize:
          mergeAllInList()

      # merge leftover files no matter the total size
      mergeAllInList()

    # return all cleaned up paths
    return _makeGlob()

  def _load(self, pattern, kind):
    result = {}
    for path in self._findPathsAndSanitize(pattern, kind):
      try:
        with open(path, 'rb') as _f:
          data = pickle.load(_f)
      except Exception as e:
        io.warn(f'failed to read {kind} file {path}: {e.__class__.__name__} "{e}"')
      else:
        # merge hitData content with result dict
        for k, v in data.items():
          _updateResultEntry(result, k, v)

    return result

class RawFolderRange:
  def __init__(self, paths):
    self._paths = [p.path() if isinstance(p, RawFolder) else p for p in paths]
  
  def __iter__(self):
    return [RawFolder(p) for p in self._paths].__iter__()

  def __len__(self):
    return len(self._paths)

  def __getitem__(self, i):
    selectedPaths = self._paths[i]
    if type(selectedPaths) in (str, str_):
      return RawFolder(selectedPaths)
    return RawFolderRange(selectedPaths)

  def paths(self):
    return [os.path.relpath(p) for p in self._paths]

  @functools.wraps(RawFolder.loadHits)
  def loadHits(self, *args, **kwargs):
    result = {}
    for res in [r.loadHits(*args, **kwargs) for r in self]:
      for k, v in res.items():
        _updateResultEntry(result, k, v)
    return hits.Hits(result)

  @functools.wraps(RawFolder.loadRays)
  def loadRays(self, *args, **kwargs):
    result = {}
    for res in [r.loadRays(*args, **kwargs) for r in self]:
      for k, v in res.items():
        _updateResultEntry(result, k, v)
    return result



__license__ = 'LGPL-3.0-or-later'
__copyright__ = 'Copyright 2024  W. Braun (epiray GmbH)'
__authors__ = 'P. Bredol'
__url__ = 'https://github.com/zaphB/freecad.optics_design_workbench'


try:
  import FreeCADGui as Gui
  import FreeCAD as App
except ImportError:
  pass

from . import find
from .. import simulation
from . common import *


class OpticalSimulationAction:
  def __init__(self, action):
    self.action = action

  def Activated(self):
    # call clear method of all light source Proxies to erase all visible rays
    if self.action == 'clear':
      for obj in find.lightSources():
        obj.Proxy.clear(obj)
        
        # make sure GUI does not freeze during this
        keepGuiResponsive()
    else:
      # forward action to simulations submodule
      simulation.runAction(self.action)

  def IsActive(self):
    return True

  def GetResources(self):
    return dict(
      Pixmap=find.iconpath(self.action),
      Accel='',
      **(dict(
        clear=dict(
          MenuText='Clear rays',
          ToolTip='Clear all displayed rays from the project.'
        ),
        fans=dict(
          MenuText='Recalculate fans',
          ToolTip='Calculate and display ray fans for all light sources.'
        ),
        singlepseudo=dict(
          MenuText='Run a single pseudo-random iteration',
          ToolTip='Run and display a single pseudo-random Monte-Carlo iteration.'
        ),
        singletrue=dict(
          MenuText='Run a single true random iteration',
          ToolTip='Run and display a single true random Monte-Carlo iteration.'
        ),
        pseudo=dict(
          MenuText='Start pseudo-random simulation',
          ToolTip='Start running the pseudo-random Monte-Carlo simulation in '
                  'the background.'
        ),
        true=dict(
          MenuText='Start true random simulation',
          ToolTip='Start running the true random Monte-Carlo simulation in '
                  'the background.'
        ),
        stop=dict(
          MenuText='Stop running simulation',
          ToolTip='Stop the simulation currently running in the background.'
        )
      )[self.action]))

def loadSimulationActions():
  Gui.addCommand('Clear all rays', OpticalSimulationAction('clear'))
  Gui.addCommand('Place ray fans', OpticalSimulationAction('fans'))
  Gui.addCommand('Single pseudo random', OpticalSimulationAction('singlepseudo'))
  Gui.addCommand('Single true random', OpticalSimulationAction('singletrue'))
  Gui.addCommand('Continuous pseudo random', OpticalSimulationAction('pseudo'))
  Gui.addCommand('Continuous true random', OpticalSimulationAction('true'))
  Gui.addCommand('Stop continuous', OpticalSimulationAction('stop'))



'''

'''

__license__ = 'LGPL-3.0-or-later'
__copyright__ = 'Copyright 2024  W. Braun (epiray GmbH)'
__authors__ = 'P. Bredol'
__url__ = 'https://github.com/zaphB/freecad.optics_design_workbench'


import functools
import traceback
import time

from .. import io

class retryOnError:
  def __init__(self, subject, maxRetries=3, callbackAfterRetries=0, callback=None):
    self.subject = subject
    self.maxRetries = maxRetries
    self.callbackAfterRetries = callbackAfterRetries
    self.callback = callback

  def __call__(self, func):
    @functools.wraps(func)
    def _wrapped(*args, _func=func, **kwargs):
      for retryNo in range(self.maxRetries+10):
        try:
          return _func(*args, **kwargs)
        except Exception:
          if retryNo >= self.maxRetries:
            raise
          elif retryNo >= self.callbackAfterRetries and self.callback:
            io.warn(f'exception raised while {self.subject}, running retry-callback and retrying... ({retryNo=}):\n\n'+traceback.format_exc())
            self.callback()
          else:
            io.warn(f'exception raised while {self.subject}, retrying... ({retryNo=}):\n\n'+traceback.format_exc())
          time.sleep(0.1)

    return _wrapped



'''

'''

__license__ = 'LGPL-3.0-or-later'
__copyright__ = 'Copyright 2024  W. Braun (epiray GmbH)'
__authors__ = 'P. Bredol'
__url__ = 'https://github.com/zaphB/freecad.optics_design_workbench'


from numpy import *
from matplotlib.pyplot import *
import seaborn as sns
import pandas as pd
import matplotlib.ticker
import seaborn as sns
import scipy.optimize

import os
import threading
import time
import traceback
import functools
import multiprocessing
import pickle
import cloudpickle
import copy
from atomicwrites import atomic_write

from .. import io
from . import freecad_document
from . import progress
from . import retries

CLOSE_FREECAD_TIMEOUT = 90
_ALL_OPEN_SWEEPERS = []


def closeAllSweepers():
  for s in _ALL_OPEN_SWEEPERS:
    s.close()


@functools.cache
def _mpCtx():
  # use safest method='spawn' even if it is rather slow, but SweeperWorkers
  # will live many minutes usually, so the overhead does not matter
  return multiprocessing.get_context('forkserver')


def _unpickleAndWork(pickledSweeperOptimizeWorker, freecadExecutable):
  '''
  wrapper around SweeperOptimizeWorker._work method that is 
  a suitable multiprocessing.Process target
  '''
  freecad_document.setDefaultFreecadExecutable(freecadExecutable)
  _self = pickle.loads(pickledSweeperOptimizeWorker)
  _self.work()


class SweeperOptimizeWorker:
  def __init__(self, sweeper, optimizeArgs):
    self._lastSentTerminate = 0
    self._lastSentKill = 0
    self._termSignalInterval = 3
    self._killSignalInterval = 3
    self._tryToEndWorkersSince = None
    self._optimizeArgs = optimizeArgs

    # set path to dump history to
    self._historyDumpPath = os.path.abspath(f'{sweeper.resultsPath()}/tmp/optimize-hist-{int(time.time()*1e3)}-{int(random.random()*1e5)}-pid{os.getpid()}-thread{threading.get_ident()}.pkl')

    # setup history dump path and randomize historyDumpInterval to
    # avoid synchronization
    self._optimizeArgs['historyDumpPath'] = self.historyDumpPath()
    self._optimizeArgs['historyDumpInterval'] = 30+30*random.random()

    # make sure background worker will never plot anything on his own
    # self._optimizeArgs['progressPlotInterval'] = inf

    # setup history attrs
    self._history = []

    # Make sure sweeper sweeper document is closed to avoid inheriting
    # the opened FreecadDocument attributes to the child process.
    # Also remove the threading lock object, which cannot be passed to 
    # the child process (lock will be recreated next time it is needed)
    self._sweeperInstance = sweeper
    self._sweeperInstance.close()
    self._sweeperInstance._freecadLock = None

    # Pickle sweeper instance and optimize args func using cloudpickle,
    # because they are usually defined in a jupyter notebook.
    # The multiprocessing module uses built-in pickle and will not be able
    # to pass functions defined in the jupyter notebook to its workers. 
    pickledSelf = cloudpickle.dumps(self)

    # setup and start child process
    self._process = _mpCtx().Process(
                        target=_unpickleAndWork, 
                        args=(pickledSelf, 
                              freecad_document._DEFAULT_FREECAD_EXECUTABLE), 
                        daemon=True ) # <- kill process after parent has exited

  def start(self):
    self._process.start()

  def work(self):
    # make sure run-in-fresh-copy-mode is enabled to prevent any worker
    # from ever touching the main FCStd file
    self._sweeperInstance.setOpenFreshCopyMode(True)

    # run optimizer work
    self._sweeperInstance.optimize(**self._optimizeArgs)

  def historyDumpPath(self):
    return self._historyDumpPath

  def fetchHistory(self):
    # try to load history to update cached history
    try:
      if os.path.exists(self.historyDumpPath()):
        with open(self.historyDumpPath(), 'rb') as f:
          self._history = pickle.load(f)
        os.remove(self.historyDumpPath())
    except Exception:      
      io.verb(f'fetching history from {self.historyDumpPath()} '
              f'failed:\n\n'+traceback.format_exc())
    # return history
    return self._history

  def isRunning(self):
    return self._process.is_alive()

  def terminate(self):
    # limit signal send frequency
    if time.time()-self._lastSentTerminate > self._termSignalInterval:
      self._lastSentTerminate = time.time()
      self._termSignalInterval += 1
      io.verb(f'sent terminate signal to {self}')
      return self._process.terminate()

  def kill(self):
    # limit signal send frequency
    if time.time()-self._lastSentKill > self._killSignalInterval:
      self._killSignalInterval += 1
      self._lastSentKill = time.time()
      io.verb(f'sent kill signal to {self}')
      return self._process.kill()

  def escalatingQuit(self):
    if self._tryToEndWorkersSince is None:
      self._tryToEndWorkersSince = time.time()
    if time.time()-self._tryToEndWorkersSince > 15:
      self.kill()
    else:
      self.terminate()


class MetaParameter:
  '''
  This class implements set and get methods like the FreeCAD document
  parameters do. The get method returns the value previously set, or 
  nan if no set command was issued so far. The set command stores the
  set value and tries to set the dependent real parameters.
  '''
  def __init__(self, name, metaParameterFunc, sweeper):
    self._metaParameterFunc = metaParameterFunc
    self._name = name
    self._sweeper = sweeper
    self._siblings = [self]
    self._latestResultDict = {}
    self._value = nan
    self._allSiblingsWereSetOnce = False

  def setSiblings(self, siblings):
    self._siblings = list(siblings)

  def set(self, value, dontApplyMetaParamYet=False, **kwargs):
    # store whether value was nan before and update our own value
    wasNan = isnan(self._value)
    self._value = value

    # check if all siblings have non-nan value
    unsetSiblings = [p._name for p in self._siblings if isnan(p.get())]
    if not len(unsetSiblings):
      # report if this is the first time we reach this point and actually use the metaParameterFunc
      if not self._allSiblingsWereSetOnce:
        io.verb(f'meta parameter family {", ".join([p._name for p in self._siblings])} is '
                f'completely initialized and active from now on')
      self._allSiblingsWereSetOnce = True

      # calculate result dict to be set according to stored metaParameterFunc
      resultDict = self._metaParameterFunc(self._sweeper, **{p._name: p.get() for p in self._siblings})

      # set result dict in all siblings, sweeper.set will internally 
      if dontApplyMetaParamYet:
        self._latestResultDict = resultDict
        for s in self._siblings:
          s._latestResultDict = resultDict

      # alternatively: directly apply new result dict
      else:
        self._sweeper.set(**resultDict)

    # if other meta param siblings have not been set yet and our previous value was
    # non-nan, issue a warning
    else:
      if not wasNan:
        io.warn(f'set meta parameter {self._name} to {value}, but its the siblings '
                f'{", ".join(unsetSiblings)} have not been set yet. Setting meta '
                f'parameters only has an effect once all siblings have been set once.')


  def get(self):
    return self._value


class ParameterSweeper:
  '''
  The parameter sweeper allows to conveniently set/get/sweep/optimize 
  parameters in the freecad files using given names, instead of the 
  lengthy descriptions in the document tree.

  Arguments:

  getParameterFunc : function
    Define how named parameters are mapped to nodes in the freecad document. 
    The function has to accept one parameter, the FreecadDocument instance and
    is expected to return a dictionary. Keys in the returned dictionary are
    the sweepable parameter names, values of the dictionary are the 
    FreecadDocument parameter nodes.
    The odd indirect definition through as a function is necessary, because
    reopening the freecad file requires to rebuilt the references to freecad
    objects.
  '''
  def __init__(self, getParametersFunc, freecadDocumentKwargs={}):
    # close all open sweepers when a new one is created to prevent
    # to make life in jupyter notebooks easier
    while len(_ALL_OPEN_SWEEPERS):
      _ALL_OPEN_SWEEPERS[0].close()

    self._getParametersFunc = getParametersFunc
    self._metaParameterDict = {}
    self._freecadDocumentKwargs = freecadDocumentKwargs
    self._freecadDocument = None
    self._closeDocumentAfterInactivityThread = None
    self._freecadLock = None
    self._bounds = {}
    self._optimizeStepsArgCache = {}
  
  def addMetaParameters(self, metaParameterFunc):
    newMetaParams = {}
    for argName in list(inspect.signature(metaParameterFunc).parameters.keys())[1:]:
      if argName in list(self.parameters().keys()):
        raise ValueError(f'meta parameter function argument {repr(argName)} '
                         f'conflicts with existing parameter. Did you already '
                         f'add these metaparameters? Or does the name exist '
                         f'among the regular parameters?')

      # add meta parameter object to the dictionary
      newMetaParams[argName] = MetaParameter(argName, metaParameterFunc, self)

    # let all newly generated meta params know about their siblings
    for v in newMetaParams.values():
      v.setSiblings(newMetaParams.values())

    # save new meta params to dictionary
    self._metaParameterDict.update(newMetaParams)
    io.verb(f'registered meta parameters {newMetaParams}')
    
    # reset self.parameters cache to make sure meta params appear on next call
    self.parameters.cache_clear()

  def _freecadDocumentLock(self):
    if self._freecadLock is None:
      self._freecadLock = threading.RLock()
    return self._freecadLock

  def _closeOnInactivity(self):
    while self._freecadDocument:
      if time.time()-self._freecadDocument.lastInteractionTime() > CLOSE_FREECAD_TIMEOUT:
        self.close()
        break
      # limit loop speed
      time.sleep(1/3)

  def save(self):
    with self._freecadDocumentLock():
      if self._freecadDocument:
        self._freecadDocument.save()

  def __del__(self):
    self.close()

  def close(self):
    with self._freecadDocumentLock():
      try:
        if self._freecadDocument:
          self._freecadDocument.close()
      except BrokenPipeError:
        pass
      self._freecadDocument = None

      # remove self from global list
      _ALL_OPEN_SWEEPERS[:] = [s for s in _ALL_OPEN_SWEEPERS if s != self]

      # clear parameter cache if file was closed
      self.parameters.cache_clear()

      # remove reference to background thread (it will end on its 
      # own soon, because we set _freecadDocument to None)
      self._closeDocumentAfterInactivityThread = None

  def setOpenFreshCopyMode(self, mode):
    # close file if mode changed to ensure it is re-opened with proper mode
    # on next occasion
    prevMode = self._freecadDocumentKwargs.get('openFreshCopy', None)
    if prevMode != mode:
      self.resultsPath.cache_clear()
      self.close()

    # update freecad document kwargs
    self._freecadDocumentKwargs['openFreshCopy'] = mode

  def getOpenFreshCopyMode(self):
    return self._freecadDocumentKwargs.get('openFreshCopy', None)

  def open(self):
    with self._freecadDocumentLock():
      if self._freecadDocument is None or not self._freecadDocument.isRunning():
        # append self to global sweeper list
        _ALL_OPEN_SWEEPERS.append(self)

        # create instance and open freecad document
        self._freecadDocument = freecad_document.FreecadDocument(**self._freecadDocumentKwargs)
        self._freecadDocument.open()

        # silence simulation progress tracker
        progress.silenceProgressTracker()

        # clear parameter cache after newly opened file
        self.parameters.cache_clear()

        # setup background thread that closes document after some inactivity
        self._closeDocumentAfterInactivityThread = threading.Thread(target=self._closeOnInactivity)
        self._closeDocumentAfterInactivityThread.start()

  def freecadDocument(self):
    with self._freecadDocumentLock():
      self.open()
      return self._freecadDocument

  @functools.cache
  def resultsPath(self):
    return self.freecadDocument().resultsPath()

  def _parameterNodeDict(self):
    with self._freecadDocumentLock():
      self.open()
      res = self._getParametersFunc(self._freecadDocument)
      res.update(self._metaParameterDict)
      if not len(res):
        raise ValueError(f'getParametersFunc return empty dict, a ParameterSweeper '
                         f'without parameters is pointless')
      return res

  @functools.cache
  def parameters(self):
    with self._freecadDocumentLock():
      return {k: v.get() for k,v in self._parameterNodeDict().items()}

  def set(self, **kwargs):
    with self._freecadDocumentLock():
      boundsDict = self.bounds()
      paramDict = self._parameterNodeDict()

      # check whether keys are valid
      for setKey in kwargs.keys():
        if setKey not in paramDict.keys():
          raise ValueError(f'parameter {setKey} does not exist in dictionary returned '
                           f'by the getParametersFunc used to create this sweeper.')

      # update parameter values
      for setKey, setVal in kwargs.items():
        # restrict set val if bounds are exceeded
        b1, b2 = boundsDict[setKey]
        if setVal < b1:
          io.warn(f'trying to set parameter {setKey} to {setVal}, which is below '
                  f'lower bound {b1}. Setting to lower bound {b1} instead.')
          setVal = b1
        if setVal > b2:
          io.warn(f'trying to set parameter {setKey} to {setVal}, which is above '
                  f'upper bound {b2}. Setting to upper bound {b2} instead.')
          setVal = b2

        # update value but dont apply meta params right away 
        paramDict[setKey].set(setVal, dontApplyMetaParamYet=True)

        # ensure value was set correctly
        success = False
        gotVal = paramDict[setKey].get()
        try:
          if isclose(setVal, gotVal, rtol=1e-3):
            success = True
        except Exception:
          success = (setVal == gotVal)
        if not success:
          raise ValueError(f'try to set parameter {setKey} to value '
                           f'{repr(setVal)}, but got value {repr(gotVal)}.')
      
      # apply all changed meta params, make sure to only apply one of each
      # sibling group to avoid many redundant calls
      appliedMetaParams = []
      for setKey, setVal in kwargs.items():
        param = paramDict[setKey]
        if isinstance(param, MetaParameter) and param not in appliedMetaParams:
          appliedMetaParams.append(param)
          appliedMetaParams.extend(param._siblings)
          self.set(**param._latestResultDict)

      # clear parameter cache if parameters were updated
      self.parameters.cache_clear()

  def setBounds(self, **kwargs):
    paramNames = self.parameters().keys()
    # make sure keys exist and bounds are well formed
    for k, v in kwargs.items():
      if k not in paramNames:
        raise ValueError(f'parameter with name {k} does not exist, '
                         f'expect one of: {", ".join(paramNames)}')
      try:
        lower, upper = sorted(list(v))
      except Exception:
        raise ValueError(f'found illegal bounds for parameter {k}: {v}, '
                         f'bounds must tuple of two numbers')
    
    # set bounds
    for k, v in kwargs.items():
      self._bounds[k] = sorted(list(v))

  def bounds(self):
    return {k: self._bounds.get(k, (-inf, inf)) for k in self.parameters().keys()}

  def optimizeStrategyBegin(self):
    self._optimizeStepsArgCache = {}
    self._optimizeStepsPosArgCache = {}

  def optimizeStrategyStep(self, *args, 
                           progressCallback=None, 
                           relWaitForParallel=None, 
                           absWaitForParallel=None, 
                           progressPlotInterval=None,
                           saveInterval=None):
    '''
    Pass one or more dictionaries with optimize args to run
    optimization steps with varied free parameters, methods, etc. in parallel

    Passing dictionaries only will cause a sequence of optimize-calls with parameters
    given by the dictionary. All dictionaries inherit keys from prededing arg-dict.
    Only changed keys need to be specified. The following example will run one optimize
    with default method and the remaining args given in the first dict. The second 
    optimize run will use the same arguments as the first, except for method='evolution'. 
    
    optimizer.optimizeStrategyStep(
      dict(method=None, minimizeFunc=...),
    )
    optimizer.optimizeStrategyStep(
      dict(method='evolution') 
    )

    If a list of dictionaries is passed instead of a dictionary, two optimize calls
    will be run in parallel. Again later dicts in the list inherit from previous dicts.
    The following example will first run the default method and a Nelder-Mead optimizer
    in parallel. After one of them finished, we will wait relWaitForParallel*runtime 
    for the other optimize run and then proceed with a method=evolution run, which itself
    will start with the best result obtained before:

    optimizer.optimizeStrategyStep(
      dict(method=None, minimizeFunc=...),
      dict(method='Nelder-Mead') 
    )
    optimizer.optimizeStrategyStep(
      dict(method='evolution') 
    )
    '''
    global CLOSE_FREECAD_TIMEOUT

    # Cache positional argument values, too. Assume Nones mean parameter was not given
    self._optimizeStepsPosArgCache.update({k:v for k,v in locals().items() if k not in ('self', 'args') and v is not None})
    progressCallback = self._optimizeStepsPosArgCache.get('progressCallback', None)
    relWaitForParallel = self._optimizeStepsPosArgCache.get('relWaitForParallel', .5)
    absWaitForParallel = self._optimizeStepsPosArgCache.get('absWaitForParallel', 300)
    progressPlotInterval = self._optimizeStepsPosArgCache.get('progressPlotInterval', 60)
    saveInterval = self._optimizeStepsPosArgCache.get('saveInterval', 5*60)

    # add cache contents to all arg dicts
    for kwargs in args:
      self._optimizeStepsArgCache.update(kwargs)
      kwargs.update(self._optimizeStepsArgCache)

    # check validity of strategy
    if not len(args):
      raise ValueError('no steps for optimization strategy given')

    # do work in this process, no workers launched
    if len(args) == 1:
      io.verb(f'running single process optimize with kwargs={args[0]}')
      self.optimize(**args[0])

    # launch workers to do work, this process just monitors
    else:
      io.verb(f'running multi process optimize with args={args}')
      t0 = time.time()
      lastProgressPlot = 0

      # increase freecad timeout duration to a very log time to avoid annoying reopenings
      # during optimize strategy
      CLOSE_FREECAD_TIMEOUT = 3600
      
      # save document and create worker objects
      self.save()
      workers = []
      io.verb(f'setting up worker processes...')
      for kwargs in args:
        workers.append(SweeperOptimizeWorker(self, kwargs))
      
      # start worker processes (do creation of workers and starting in separate loops to avoid
      # stretching the period of worker launch induced window flickering for too long)
      io.verb(f'launching worker processes...')
      for w in workers:
        # sleep random time between worker creation to avoid stressing the filesystem unnecessarily
        time.sleep(.2+.2*random.random())
        w.start()

      bestParamsDict = None
      bestParamsArgs = None
      try:
        # monitor workers until happy
        lastWorkerFinished = inf
        runningWorkers = list(workers)
        bestPenalty = inf
        lastPenaltyImprovement = 0
        tryToEndWorkersSince = inf
        lastDocumentSave = time.time()
        while True:

          # fetch history of all worker progresses
          allParamsHist = []
          for w in workers:
            allParamsHist.extend(w.fetchHistory())
          allParamsHist = sorted(allParamsHist, key=lambda e: e[0])
          while len(allParamsHist) > 1e4:
            allParamsHist = allParamsHist[::2]

          # check if global best-penalty improved
          if len(allParamsHist) and (_newBest:=min([h[1] for h in allParamsHist])) < bestPenalty:
            bestPenalty = _newBest
            lastPenaltyImprovement = time.time()
            _best = allParamsHist[argmin([h[1] for h in allParamsHist])]
            bestParamsDict = _best[4]
            bestParamsArgs = _best[5]
            io.verb(f'found new best solution {bestPenalty=},\n{bestParamsDict=}\n{bestParamsArgs=}')
            _b = self.bounds()
            _paramsRelToBounds = {k: (v-_b[k][0])/(_b[k][1]-_b[k][0]) 
                                           for k,v in bestParamsDict.items()}
            io.verb(f'params in best solution that are close to bounds: '
                    f'{[k for k,v in _paramsRelToBounds.items() if isclose(v, 0, atol=1e-3) or isclose(v, 1, atol=1e-3)]} '
                    f'(all params renormalized to bounds: {_paramsRelToBounds})')

          # update non-temp document every now and then with best params so far and 
          # save to disk to avoid losing all on a crash
          if time.time()-lastDocumentSave > saveInterval:
            lastDocumentSave = time.time()
            io.verb('autosaving current best result')
            self.set(**bestParamsDict)
            self.save()

          # plot history of optimization and hits of best result so far
          if len(allParamsHist) > 15 and time.time()-lastProgressPlot > progressPlotInterval:
            lastProgressPlot = time.time()
            progress.clearCellOutput()

            fig, ax1 = subplots(1, 1, figsize=(6,4))
            sca(ax1)
            sns.scatterplot(pd.DataFrame([p[:3] for p in allParamsHist]), x=0, y=1, 
                            style=2, size=2, markers=['.', '*'], sizes=[15, 40], legend=False,
                                        ).set(xlabel='time', ylabel='penalty')
            _allFinitePenalties = [p[1] for p in allParamsHist if isfinite(p[1])]
            if len(_allFinitePenalties) > 50:
              l, u = min(_allFinitePenalties), quantile(_allFinitePenalties, .5)
              if min(_allFinitePenalties) > 0 and u/l > 30:
                ax1.semilogy()
                ax1.set_ylim([l / (u/l)**0.05, u * (u/l)**0.5])
              else:
                ax1.set_ylim([l-.05*(u-l), u+0.5*(u-l)])
            ax1.xaxis.set_major_formatter(matplotlib.ticker.FuncFormatter(
                                              lambda x, p: io.secondsToStr(x-t0, length=1) ))
            ax1.set_title(f'penalty history ({len(runningWorkers)}/{len(workers)} workers busy)', fontsize=10)

            # save plot to disk
            savefig(f'{self.resultsPath()}/optimize-progress.pdf')

            # show plot in notebook
            show()

            # close figure
            close()

            # print status
            io.info(f'optimize strategy step running since {io.secondsToStr(time.time()-t0)}, {len(runningWorkers)}/{len(workers)} workers busy')

            # run custom progress callback if given
            if progressCallback:
              progressCallback(bestParams=bestParamsDict, history=allParamsHist)
            lastProgressPlot = time.time()

          # update running workers list
          for w in runningWorkers:
            if not w.isRunning():
              io.verb(f'worker {w} finished')
              lastWorkerFinished = time.time()
          runningWorkers = [w for w in runningWorkers if w.isRunning()]

          # end loop if all workers finished
          if not len(runningWorkers):
            io.verb(f'all workers finished, exiting...')
            break

          # if at least one worker finished and none of the other workers managed
          # to improve the penalty since relWaitForParallel*runtime, exit all remaining
          # workers
          if ( not isfinite(tryToEndWorkersSince)
               and time.time()-lastWorkerFinished > relWaitForParallel*(lastWorkerFinished-t0)
                                                      + absWaitForParallel
               and time.time()-lastPenaltyImprovement > relWaitForParallel*(lastWorkerFinished-t0)
                                                          + absWaitForParallel ):
            io.verb(f'at least one worker finished '
                    f'({io.secondsToStr(time.time()-lastWorkerFinished)} ago) '
                    f'and others did not improve for more '
                    f'than {io.secondsToStr(relWaitForParallel*(lastWorkerFinished-t0))}, '
                    f'(last improvement {io.secondsToStr(time.time()-lastPenaltyImprovement)} ago) '
                    f'quitting remaining workers...')
            tryToEndWorkersSince = time.time()
          
          # send kill/terminate signals depending on wait time
          if time.time()-tryToEndWorkersSince > 0:
            for w in runningWorkers:
              w.escalatingQuit()

          # limit loop speed
          time.sleep(3)
        
      # make sure to apply best result to current FCStd file if loop ends
      finally:
        io.info(f'optimize strategy step ended, {bestParamsDict=}')
        if bestParamsDict:
          self.set(**bestParamsDict)
          self.save()

        # wait for all workers to finish
        lastPrint = time.time()
        while True:
          runningWorkers = [w for w in workers if w.isRunning()]
          if not len(runningWorkers):
            break
          if time.time()-lastPrint > 10:
            lastPrint = time.time()
            io.warn(f'optimize strategy step ended, but still waiting for {len(runningWorkers)} workers to exit...')

          # send kill/terminate signals depending on wait time
          for w in runningWorkers:
            w.escalatingQuit()

          # limit loop speed
          time.sleep(1/2)

        # make sure all progress files are cleared
        for w in workers:
          w.fetchHistory()

        # restore standard 90s freecad timeout
        CLOSE_FREECAD_TIMEOUT = 90

  def optimizeStrategyEnd(self):
    self._optimizeStepsArgCache = {}
    self.purgeTempFolder()
  

  def purgeTempFolder(self):
    self.freecadDocument().purgeTempFolder()


  def optimize(self, minimizeFunc, parameters, simulationMode, 
               prepareSimulation=None, simulationKwargs={},
               minimizerKwargs={}, progressPlotInterval=30, 
               method=None, historyDumpPath=None, 
               historyDumpInterval=inf, **kwargs):
    # save optimize params to variable
    optimizeParams = {k:v for k,v in locals().items() if k not in ('self',)}

    # setup progress and timing vars
    t0 = time.time()
    lastProgressPlot = 0
    lastHistoryDump = time.time()
    minimizeFuncHist = []
    allParamsHist = []
    bestPenaltySoFar, bestParametersSoFar, bestResultSoFar = inf, None, None

    parameters = list(parameters)
    with self._freecadDocumentLock():
      self.open()

      # wrap minimize func, run simulation before and pass additional args
      def _simulateAndCalcMinimizeFunc(args):
        nonlocal lastProgressPlot, lastHistoryDump
        nonlocal bestPenaltySoFar, bestParametersSoFar, bestResultSoFar

        # to enhance stability: if something raises an error during the function
        # evaluation, return a very large number
        try:
          # run prepare simulation-hook
          @retries.retryOnError(subject='preparing simulation')
          def _prepareSimulation():
            if prepareSimulation:
              prepareSimulation(self._freecadDocument, **kwargs)
          _prepareSimulation()

          # extract param dict from args, un-normalize parameters that have both bounds set
          _b = self.bounds()
          paramDict = {k: v*(_b[k][1]-_b[k][0])+_b[k][0] if all(isfinite(_b[k])) else v 
                                                          for k,v in zip(parameters, args)}

          # run simulation, if simulating fails, set penalty to very large number
          @retries.retryOnError(subject='setting parameters and running simulation',
                                maxRetries=4, callbackAfterRetries=2, callback=self.close)
          def _runSimulation():
            self.set(**paramDict)
            return self._freecadDocument.runSimulation(simulationMode, **simulationKwargs)
          resultFolder = _runSimulation()

          # plot progress if it is time (do this before the call to minimize func to make sure 
          # any output of minimize func will be visible below the progress info)
          if time.time()-lastProgressPlot > progressPlotInterval and len(allParamsHist) > 5:
            lastProgressPlot = time.time()
            progress.clearCellOutput()

            # plot history of optimization and hits of best result so far
            fig, ax1 = subplots(1, 1, figsize=(6,4))
            sca(ax1)
            sns.scatterplot(pd.DataFrame([p[:3] for p in allParamsHist]), x=0, y=1, 
                            style=2, size=2, markers=['.', '*'], sizes=[15, 40], legend=False,
                                        ).set(xlabel='time', ylabel='penalty')
            gca().xaxis.set_major_formatter(matplotlib.ticker.FuncFormatter(
                                              lambda x, p: io.secondsToStr(x-t0, length=1) ))
            gca().set_title(f'penalty history', fontsize=10) 

            # save plot to disk
            savefig(f'{self.resultsPath()}/optimize-progress.pdf')

            # show in notebook (if not running as worker)
            if not self.getOpenFreshCopyMode():
              show()

            # close figure
            close()

            # print status
            io.info(f'optimizer running since {io.secondsToStr(time.time()-t0)}')

          # calculate penalty
          @retries.retryOnError(subject='evaluating minimize func')
          def _calcPenalty():
            return minimizeFunc(resultFolder)
          penalty = _calcPenalty()

          # update history lists and shorten if necessary
          if penalty < bestPenaltySoFar:
            io.verb(f'found new optimum: {penalty=}, {paramDict=}')
            allParamsHist.append([time.time(), penalty, True, 
                                  os.path.realpath(resultFolder.path()), paramDict, 
                                  optimizeParams])
            bestParametersSoFar = dict(paramDict)
            bestPenaltySoFar = penalty
            bestResultSoFar = resultFolder
            _b = self.bounds()
            _paramsRelToBounds = {k: (v-_b[k][0])/(_b[k][1]-_b[k][0]) 
                                            for k,v in paramDict.items()}
            io.verb(f'params in best solution that are close to bounds: '
                    f'{[k for k,v in _paramsRelToBounds.items() if isclose(v, 0, atol=1e-3) or isclose(v, 1, atol=1e-3)]} '
                    f'(all params renormalized to bounds: {_paramsRelToBounds})')
          else:
            allParamsHist.append([time.time(), penalty, False, 
                                  os.path.realpath(resultFolder.path()), paramDict, 
                                  optimizeParams])
          while len(allParamsHist) > 1e4:
            allParamsHist[:] = allParamsHist[::2]
          
          # dump entire history to file if enabled
          if historyDumpPath:
            if time.time()-lastHistoryDump > historyDumpInterval:
              lastHistoryDump = time.time()
              try:
                os.makedirs(os.path.dirname(historyDumpPath), exist_ok=True)
                with atomic_write(historyDumpPath, mode='wb', overwrite=True) as f:
                  cloudpickle.dump(allParamsHist, f)
              except Exception:
                io.warn(f'dumping progress failed:\n\n'+traceback.format_exc())

          # return the penalty value
          return penalty

        # capture any exception, log the stack trace and return ridiculously large number
        except Exception:
          io.warn(f'exception was raised in optimizer iteration:\n\n'+traceback.format_exc())
          return 1e99

      # prepare arguments for minimizer: if params have both limits set, renormalize to (0,1) interval
      _b = self.bounds()
      _p = self.parameters()
      x0 = [(_p[k]-_b[k][0])/(_b[k][1]-_b[k][0]) if all(isfinite(_b[k])) else _p[k] for k in parameters]
      bounds = [[-.001,1.001] if all(isfinite(_b[k])) else _b[k] for k in parameters]
      io.info(f'starting optimizer with {method=} {minimizeFunc=}, {parameters=}, {simulationMode=}, '
              f'{simulationKwargs=}, {kwargs=}, {x0=}, {bounds=}')

      # run actual minimizer
      try:
        if method == 'annealing':
          return scipy.optimize.dual_annealing(_simulateAndCalcMinimizeFunc, x0=x0, bounds=bounds, **minimizerKwargs) 
        elif method == 'evolution':
          return scipy.optimize.differential_evolution(_simulateAndCalcMinimizeFunc, x0=x0, bounds=bounds, **minimizerKwargs) 
        if method:
          minimizerKwargs['method'] = method
        return scipy.optimize.minimize(_simulateAndCalcMinimizeFunc, x0=x0, bounds=bounds, **minimizerKwargs)

      # before returning, make sure parameters for global optimum are set
      finally:
        if bestParametersSoFar:
          self.set(**bestParametersSoFar)
          self.save()



'''

'''

__license__ = 'LGPL-3.0-or-later'
__copyright__ = 'Copyright 2024  W. Braun (epiray GmbH)'
__authors__ = 'P. Bredol'
__url__ = 'https://github.com/zaphB/freecad.optics_design_workbench'


from numpy import *

import threading
import time
import os
import pickle
import traceback

from .. import io

try:
  import IPython.display
  hasIPython = True
except ImportError:
  hasIPython = False

ALLOW_PROGRESS_TACKERS = False
_GLOBAL_PROGRESS_TRACKER = None

def setupProgressTracker(**kwargs):
  '''
  Create a new global progress tracker for usage in jupyter notebooks
  '''
  global _GLOBAL_PROGRESS_TRACKER

  # raise if progress tracker creation is not allowed (=we are not in a FreecadDocument context)
  if not ALLOW_PROGRESS_TACKERS:
    raise ValueError(f'progress tracking can only be setup within FreecadDocument(..) contexts')

  # quit existing progress tracker if any
  if _GLOBAL_PROGRESS_TRACKER is not None and not _GLOBAL_PROGRESS_TRACKER._isQuit:
    _GLOBAL_PROGRESS_TRACKER.quit()
  
  # make sure new progress tracker is silent if older one was silent
  if _GLOBAL_PROGRESS_TRACKER is not None and _GLOBAL_PROGRESS_TRACKER._silent:
    kwargs.update(dict(silent=True))

  # create new progress tracker
  _GLOBAL_PROGRESS_TRACKER = _ProgressTacker(**kwargs)

  # return global instance
  return _GLOBAL_PROGRESS_TRACKER

def silenceProgressTracker():
  setupProgressTracker(silent=True)

def progressTrackerExists():
  return _GLOBAL_PROGRESS_TRACKER is not None and not _GLOBAL_PROGRESS_TRACKER._isQuit

def progressTrackerInstance(**kwargs):
  '''
  Fetch the current global progressTrackerInstance
  '''
  if _GLOBAL_PROGRESS_TRACKER is None or _GLOBAL_PROGRESS_TRACKER._isQuit:
    setupProgressTracker(**kwargs)
  return _GLOBAL_PROGRESS_TRACKER

def clearCellOutput():
  IPython.display.clear_output(True)


class _ProgressTacker:
  def __init__(self, doc=None, totalSimulations=None, silent=False):
    self._doc = doc
    self._simulationNo = 0
    self._totalSimulations = totalSimulations
    self._isRunning = True
    self._isQuit = False
    self._clearCallCount = 0
    self._t = threading.Thread(target=self.updateLoop)
    self._t0 = time.time()
    self._previousProgressDict = None
    self._silent = silent
    self.resultsFolder = None
    self.start()

  def _clear(self):
    # Do not clear output on the first calls to clear, because this would
    # erase exception stacktraces that may be raised in the very first few
    # iterations of a simulation loop in the jupyter cell. 
    self._clearCallCount += 1
    if not self._silent and hasIPython and self._clearCallCount > 5:
      IPython.display.clear_output(True)

  def start(self):
    self._clear()
    #print('setting up simulation progress tracking...')
    self._t.start()
    self._t0 = time.time()

  def update(self, displayTiming=True):
    if self._silent:
      return

    if self.resultsFolder:
      try:
        latest = sorted([f for f in os.listdir(f'{self.resultsFolder._path}/progress')
                                                      if f.startswith('master') ])[-1]
        with open(f'{self.resultsFolder._path}/progress/{latest}', 'rb') as _f:
          p = pickle.load(_f)
        self._previousProgressDict = p
      except (FileNotFoundError, IndexError):
        p = self._previousProgressDict or dict(
                  totalIterations=nan, endAfterIterations=nan,
                  totalRecordedHits=nan, endAfterHits=nan,
                  totalTracedRays=nan, endAfterRays=nan)
      except Exception:
        io.warn(traceback.format_exc())

      # calculate elapsed and expected remaining time
      elapsed = time.time()-self._t0
      simulationProg = max([ p['totalIterations']/p['endAfterIterations'],
                              p['totalRecordedHits']/p['endAfterHits'],
                              p['totalTracedRays']/p['endAfterRays'] ])
      if not isfinite(simulationProg):
        simulationProg = 0
      relProgress = (self._simulationNo + simulationProg)/(self._totalSimulations or 1)
      expectedRemain = inf
      if relProgress > 0:
        expectedRemain = elapsed/relProgress * (1-relProgress)
      if expectedRemain > elapsed**2:
        expectedRemain = None

      # generate progress message
      self._clear()
      iterationProgress = ''
      if self._totalSimulations:
        iterationProgress = f'simulations done {self._simulationNo}/{self._totalSimulations}'
      simulationProgress = ''
      if isfinite(p['totalIterations']):
        simulationProgress = (f'current simulation: '
                              f"iter {p['totalIterations']}/{p['endAfterIterations']}, "
                              f"hits {p['totalRecordedHits']}/{p['endAfterHits']}, "
                              f"rays {p['totalTracedRays']}/{p['endAfterRays']}")
      message = ', '.join([s for s in (iterationProgress, simulationProgress) if s.strip()])

      # print progress message
      printedSomething = False
      if message.strip():
        print(message)
        printedSomething = True
      if displayTiming:
        print(f'elapsed time {io.secondsToStr(elapsed)}'
              +(f', expected remaining time {io.secondsToStr(expectedRemain)}' 
                                        if expectedRemain is not None else ''))
        printedSomething = True

      # return whether something was printed or not
      return printedSomething

  def updateLoop(self):
    while self._isRunning:
      self.update()
      time.sleep(1/3)
    if self.update(displayTiming=False):
      print(f'simulations ended after {io.secondsToStr(time.time()-self._t0)}')

  def nextIteration(self):
    self._simulationNo += 1

    # reset buffered progress dict to not display progress of previous iteration
    self._previousProgressDict = None

    # if total iterations was never specified, quit 
    # this progress tracker such that a new one is
    # created in the next iteration (if any)
    if self._totalSimulations is None:
      self.quit()
  
  def quit(self):
    self._isRunning = False
    self._t.join()
    self._isQuit = True
